#!/usr/bin/env python
import json
from json import JSONDecodeError
import pathlib
import re
import sys
import unittest
from unittest import mock
from unittest.mock import ANY

from PySide6 import QtCore, QtGui, QtTest, QtWidgets

from test.asammdf.gui.widgets.test_BasePlotWidget import TestPlotWidget


class TestContextMenu(TestPlotWidget):
    # Note: Test Plot Widget through FileWidget.

    def setUp(self):
        super().setUp()

        self.setUpFileWidget(measurement_file=self.measurement_file, default=True)
        # Switch ComboBox to "Natural sort"
        self.widget.channel_view.setCurrentText("Natural sort")

        self.create_window(window_type="Plot", channels_indexes=[10, 11, 12, 13, 15])
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)

        # Drag and Drop channel from FileWidget.channel_tree to Plot
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()
        self.plot_channel_a = self.plot.channel_selection.topLevelItem(0)
        self.plot_channel_a.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_b = self.plot.channel_selection.topLevelItem(1)
        self.plot_channel_b.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_c = self.plot.channel_selection.topLevelItem(2)
        self.plot_channel_c.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_d = self.plot.channel_selection.topLevelItem(3)
        self.plot_channel_d.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_e = self.plot.channel_selection.topLevelItem(4)
        self.plot_channel_e.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)

        self.processEvents()

    def test_Action_SearchItem_Cancel(self):
        """
        Test Scope:
            Ensure that action 'Search Item' call 'QtWidgets.QInputDialog.getText' in order to get a channel name.
            The Dialog is canceled and nothing happens.
        Events:
            - Open Context Menu
            - Trigger 'Search Item' action
                - Simulate that the dialog is canceled.
        Evaluate:
            - Ensure that there is no selected channel because
            performing Right Click for ContextMenu will clear selection of the tree.
        """
        QtTest.QTest.keyClick(self.plot.channel_selection.viewport(), QtCore.Qt.Key.Key_Down)
        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = None, None
            self.context_menu(action_text="Search item")

        self.assertEqual(0, len(self.plot.channel_selection.selectedItems()))

    def test_Action_SearchItem_NonexistentChannel(self):
        """
        Test Scope:
            Ensure that action 'Search Item' call 'QtWidgets.QInputDialog.getText' in order to get a channel name.
            The Dialog is filled with non-existing channel name.
        Events:
            - Open Context Menu
            - Trigger 'Search Item' action
                - Simulate that the dialog is filled with non-existing channel name.
        Evaluate:
            - Ensure that selection is cleared.
            - Ensure that MessageBox informs user.
        """
        QtTest.QTest.keyClick(self.plot.channel_selection.viewport(), QtCore.Qt.Key.Key_Down)
        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))

        with (
            mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText,
            mock.patch("asammdf.gui.widgets.tree.MessageBox.warning") as mo_warning,
        ):
            mo_getText.return_value = self.id(), True
            self.context_menu(action_text="Search item")

        self.assertEqual(0, len(self.plot.channel_selection.selectedItems()))
        mo_warning.assert_called_with(self.plot.channel_selection, "No matches found", ANY)

    def test_Action_SearchItem_ExistentChannel(self):
        """
        Test Scope:
            Ensure that action 'Search Item' call 'QtWidgets.QInputDialog.getText' in order to get a channel name.
            The Dialog is filled with existing channel name.
            Found channel is selected on tree
        Events:
            - Open Context Menu
            - Trigger 'Search Item' action
                - Simulate that the dialog is filled with existing channel name.
        Evaluate:
            - Ensure that selection is set on the found channel.
        """
        QtTest.QTest.keyClick(self.plot.channel_selection.viewport(), QtCore.Qt.Key.Key_Down)
        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))

        with (
            mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText,
            mock.patch("asammdf.gui.widgets.tree.MessageBox.warning") as mo_warning,
        ):
            mo_getText.return_value = self.plot_channel_b.text(self.Column.NAME), True
            self.context_menu(action_text="Search item")

        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))
        mo_warning.assert_not_called()

    def test_Action_AddChannelGroup_Cancel(self):
        """
        Test Scope:
            - Ensure that action 'Add channel group' will add new item (type Group) on tree.
        Events:
            - Open Context Menu
            - Trigger 'Add channel group' action
                - Simulate that the dialog is canceled.
        """
        channels_nr = self.plot.channel_selection.topLevelItemCount()

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", None
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        self.assertEqual(channels_nr, self.plot.channel_selection.topLevelItemCount())

    def test_Action_AddChannelGroup(self):
        """
        Test Scope:
            - Ensure that action 'Add channel group' will add new item (type Group) on tree.
        Events:
            - Open Context Menu
            - Trigger 'Add channel group' action
                - Simulate that the dialog is confirmed.
        Evaluate:
            - Check if there is one extra element in tree.
        """
        channels_nr = self.plot.channel_selection.topLevelItemCount()

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        self.assertEqual(channels_nr + 1, self.plot.channel_selection.topLevelItemCount())

    def test_Action_CopyNames(self):
        """
        Test Scope:
            - Ensure that action 'Copy Names' will place channel names in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy Names' action
            - Select two channels
            - Open Context Menu
            - Trigger 'Copy Names' action
        Evaluate:
            - Evaluate that channels names are placed in clipboard.
        """

        with self.subTest("1Channel"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Copy names [Ctrl+N]", position=position)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            self.assertEqual(self.plot_channel_a.text(self.Column.NAME), clipboard)

        with self.subTest("2Channels"):
            self.plot_channel_a.setSelected(True)
            self.plot_channel_b.setSelected(True)
            position_1 = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            self.context_menu(action_text="Copy names [Ctrl+N]", position=position_1)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            channels = (self.plot_channel_a.text(self.Column.NAME), self.plot_channel_b.text(self.Column.NAME))
            self.assertEqual("\n".join(channels), clipboard)

    def test_Action_CopyNamesAndValues(self):
        """
        Test Scope:
            - Ensure that action 'Copy names and value' will place channel names in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy names and value' action
            - Select two channels
            - Open Context Menu
            - Trigger 'Copy names and value' action
        Evaluate:
            - Evaluate that channels names are placed in clipboard.
        """

        with self.subTest("1Channel"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Copy names and values", position=position)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            pattern_name = re.escape(self.plot_channel_a.text(self.Column.NAME))
            pattern_unit = re.escape(self.plot_channel_a.text(self.Column.UNIT))
            self.assertRegex(clipboard, expected_regex=f"{pattern_name}, t = \\d+[.]?\\d+s, \\d+[.]?\\d+{pattern_unit}")

        with self.subTest("2Channels"):
            self.plot_channel_a.setSelected(True)
            self.plot_channel_b.setSelected(True)
            position_1 = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            self.context_menu(action_text="Copy names and values", position=position_1)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            expected_regex = []
            for channel in (self.plot_channel_a, self.plot_channel_b):
                pattern_name = re.escape(channel.text(self.Column.NAME))
                pattern_unit = re.escape(channel.text(self.Column.UNIT))
                expected_regex.append(f"{pattern_name}, t = \\d+[.]?\\d+s, \\d+[.]?\\d+{pattern_unit}")

            self.assertRegex(clipboard, "\\n".join(expected_regex))

    def test_Action_CopyDisplayProperties_Channel(self):
        """
        Test Scope:
            - Ensure that channel display properties are copied in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that channel display properties are stored in clipboard in json format.
        """
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Copy display properties [Ctrl+Shift+C]", position=position)

        clipboard = QtWidgets.QApplication.instance().clipboard().text()
        try:
            content = json.loads(clipboard)
        except JSONDecodeError:
            self.fail("Clipboard Content cannot be decoded as JSON content.")
        else:
            self.assertIsInstance(content, dict)

    def test_Action_PasteDisplayProperties_Channel(self):
        """
        Test Scope:
            - Ensure that channel display properties can be pasted over different channel.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Select a different channel
            - Open Context Menu
            - Trigger 'Paste display properties' action
            - Select dst channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that display properties are transferred from one channel to another
        """
        action_copy = "Copy display properties [Ctrl+Shift+C]"
        action_paste = "Paste display properties [Ctrl+Shift+V]"

        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text=action_copy, position=position_src)

        channel_a_properties = QtWidgets.QApplication.instance().clipboard().text()

        # Paste
        position_dst = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text=action_paste, position=position_dst)

        # Copy
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text=action_copy, position=position_src)

        channel_b_properties = QtWidgets.QApplication.instance().clipboard().text()

        self.assertEqual(channel_a_properties, channel_b_properties)

    def test_Action_CopyChannelStructure_Channel(self):
        """
        Test Scope:
            - Ensure that channel structure is copied in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy channel structure' action
        Evaluate:
            - Evaluate that channel structure is stored in clipboard in json format.
        """
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Copy channel structure [Ctrl+C]", position=position)

        clipboard = QtWidgets.QApplication.instance().clipboard().text()
        try:
            content = json.loads(clipboard)
        except JSONDecodeError:
            self.fail("Clipboard Content cannot be decoded as JSON content.")
        else:
            self.assertIsInstance(content, list)
            for channel_properties in content:
                self.assertIsInstance(channel_properties, dict)

    def test_Action_PasteChannelStructure_Channel(self):
        """
        Test Scope:
            - Ensure that channel and structure is duplicated and structure is kept.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Trigger 'Copy channel structure' action
            - Open Context Menu
            - Trigger 'Paste channel structure' action
            - Select dst channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that channel is duplicated and structure is kept.
        """
        action_copy_dsp_properties = "Copy display properties [Ctrl+Shift+C]"
        action_copy = "Copy channel structure [Ctrl+C]"
        action_paste = "Paste channel structure [Ctrl+V]"

        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        # Copy Channel Structure
        self.context_menu(action_text=action_copy, position=position_src)

        channels_count = self.plot.channel_selection.topLevelItemCount()

        # Paste Channel Structure
        self.context_menu(action_text=action_paste)

        self.assertEqual(channels_count + 1, self.plot.channel_selection.topLevelItemCount())

        channels = self.plot.channel_selection.findItems(
            self.plot_channel_a.text(self.Column.NAME), QtCore.Qt.MatchFlags()
        )
        self.assertEqual(2, len(channels))

        # Copy DSP Properties
        channel_properties = []
        for channel in channels:
            position_src = self.plot.channel_selection.visualItemRect(channel).center()
            self.context_menu(action_text=action_copy_dsp_properties, position=position_src)
            channel_properties.append(QtWidgets.QApplication.instance().clipboard().text())

        self.assertEqual(channel_properties[0], channel_properties[1])

        self.processEvents(0.2)

    def test_Menu_EnableDisable_Action_DeactivateGroups_Channel(self):
        """
        Test Scope:
            - Ensure that group items can be disabled from Context Menu.
        Events:
            - Select 1 channel
            - Open Context Menu
            - Trigger "Enable/Disable -> Deactivate Groups" action
        Expected:
            - Individual channels that are not part of group should not be affected
        """
        # Events
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Deactivate groups", position=position)

        # Evaluate
        self.assertEqual(False, self.plot_channel_a.isDisabled())

    def test_Menu_EnableDisable_Action_DisableAllButThis(self):
        """
        Test Scope:
            - Ensure that all channels are disabled except the item selected.
        Events:
            - Open Context Menu
            - Select action "Disable all but this" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that all channels are disabled except selected item.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        positions_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text="Disable all but this", position=positions_src)

        # Evaluate
        self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.NAME))
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info and item != self.plot_channel_b:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))

    def test_Menu_ShowHide_Action_HideDisabledItems(self):
        """
        Test Scope:
            - Ensure that item is hidden from channel selection when is disabled.
        Events:
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that items that are unchecked are not present anymore on channel selection
        """
        self.context_menu(action_text="Hide disabled items")

        with self.subTest("DisableBySpace"):
            # Select one channel
            self.mouseClick_WidgetItem(self.plot_channel_a)
            # Event
            QtTest.QTest.keyClick(self.plot.channel_selection, QtCore.Qt.Key.Key_Space)
            # Evaluate
            self.assertTrue(self.plot_channel_a.isHidden())

        with self.subTest("DisableByClick"):
            pos = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            # Magic Number to detect center of checkbox
            pos = QtCore.QPoint(28, pos.y())
            # Event
            QtTest.QTest.mouseClick(
                self.plot.channel_selection.viewport(),
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                pos,
            )
            # Evaluate
            self.assertTrue(self.plot_channel_b.isHidden())

    def test_Menu_ShowHide_Action_ShowDisabledItems(self):
        """
        Test Scope:
            - Ensure that item is showed on channel selection when is disabled.
        Events:
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that items that are unchecked are not present anymore on channel selection
        """

        with self.subTest("DisableBySpace"):
            self.context_menu(action_text="Hide disabled items")
            # Select one channel
            self.mouseClick_WidgetItem(self.plot_channel_a)
            # Event
            QtTest.QTest.keyClick(self.plot.channel_selection, QtCore.Qt.Key.Key_Space)
            # Evaluate
            self.assertTrue(self.plot_channel_a.isHidden())
            self.context_menu(action_text="Hide disabled items")
            self.assertFalse(self.plot_channel_a.isHidden())

        with self.subTest("DisableByClick"):
            self.context_menu(action_text="Hide disabled items")
            pos = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            # Magic Number to detect center of checkbox
            pos = QtCore.QPoint(28, pos.y())
            # Event
            QtTest.QTest.mouseClick(
                self.plot.channel_selection.viewport(),
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                pos,
            )
            # Evaluate
            self.assertTrue(self.plot_channel_b.isHidden())
            self.context_menu(action_text="Hide disabled items")
            self.assertFalse(self.plot_channel_b.isHidden())

    def test_Menu_ShowHide_Action_HideMissingItems(self):
        """
        Test Scope:
            - Ensure that missing item is hidden from channel selection.
        Events:
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that missing items are not present anymore on channel selection
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        plot_channel = self.find_channel(channel_tree=self.plot.channel_selection, channel_name="1stMissingItem")
        self.processEvents()

        self.assertFalse(plot_channel.isHidden())

        # Events
        self.context_menu(action_text="Hide missing items")

        # Evaluate
        self.assertTrue(plot_channel.isHidden())
        self.processEvents(timeout=0.01)

    def test_Menu_ShowHide_Action_ShowMissingItems(self):
        """
        Test Scope:
            - Ensure that missing item is visible from channel selection.
        Events:
            - Open Context Menu
            - Select action: Hide missing items
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that missing items are present anymore on channel selection
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        self.context_menu(action_text="Hide missing items")

        plot_channel = self.find_channel(channel_tree=self.plot.channel_selection, channel_name="1stMissingItem")
        self.processEvents()

        self.assertTrue(plot_channel.isHidden())

        # Events
        self.context_menu(action_text="Hide missing items")

        # Evaluate
        self.assertFalse(plot_channel.isHidden())
        self.processEvents(timeout=0.01)

    def test_Menu_ShowHide_Action_FilterOnlyComputedChannels(self):
        """
        Test Scope:
            - Ensure that all channels are hidden from channel selection except VirtualChannels.
        Events:
            - Open Context Menu
            - Select Filter Only Computed Channels
        Evaluate:
            - Evaluate that channels items are not present anymore on channel selection except VirtualChannels.
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item:
                self.assertFalse(item.isHidden())
            iterator += 1

        # Events
        self.context_menu(action_text="Filter only computed channels")

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item and item.text(0) == "FirstVirtualChannel":
                self.assertFalse(item.isHidden())
            else:
                self.assertTrue(item.isHidden())
            iterator += 1

    def test_Menu_ShowHide_Action_UnfilterComputedChannels(self):
        """
        Test Scope:
            - Ensure that all channels are hidden from channel selection except VirtualChannels.
        Events:
            - Open Context Menu
            - Select Filter Only Computed Channels
        Evaluate:
            - Evaluate that channels items are not present anymore on channel selection except VirtualChannels.
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        # Events
        self.context_menu(action_text="Filter only computed channels")
        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item and item.text(0) == "FirstVirtualChannel":
                self.assertFalse(item.isHidden())
            else:
                self.assertTrue(item.isHidden())
            iterator += 1

        # Events
        self.context_menu(action_text="Filter only computed channels")

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item:
                self.assertFalse(item.isHidden())
            iterator += 1

    def test_Action_EditYAxisScaling(self):
        """
        Test Scope:
            - Ensure that action forwards the request to plot.
        Event:
            - Select channel
            - Open Context Menu
            - Select 'Edit Y axis scaling'
        Evaluate:
            - Evaluate that event is forwarded to plot
        """
        # Setup
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        with mock.patch.object(self.plot, "keyPressEvent") as mo_keyPressEvent:
            self.context_menu(action_text="Edit Y axis scaling [Ctrl+G]", position=position)
            mo_keyPressEvent.assert_called()
            event = mo_keyPressEvent.call_args.args[0]
            self.assertEqual(QtCore.QEvent.Type.KeyPress, event.type())
            self.assertEqual(QtCore.Qt.Key.Key_G, event.key())
            self.assertEqual(QtCore.Qt.KeyboardModifier.ControlModifier, event.modifiers())

    def test_Action_AddToCommonYAxis(self):
        """
        Test Scope:
            - Ensure that action will mark as checked checkbox for Y Axis.
        Event:
            - Select one channel
            - Open Context Menu
            - Select 'Add to common Y axis'
            - Select two channels
            - Open Context Menu
            - Select 'Add to common Y axis'
        Evaluate:
            - Evaluate that checkbox on column COMMON_AXIS is checked.
        """
        with self.subTest("1Channel"):
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

        with self.subTest("2Channels"):
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            position = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

    def test_Action_RemoveFromCommonYAxis(self):
        """
        Test Scope:
            - Ensure that action will mark as unchecked checkbox for Y Axis.
        Event:
            - Select one channel
            - Open Context Menu
            - Select 'Remove from common Y axis'
            - Select two channels
            - Open Context Menu
            - Select 'Remove from common Y axis'
        Evaluate:
            - Evaluate that checkbox on column COMMON_AXIS is unchecked.
        """
        with self.subTest("1Channel"):
            # Setup
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()

            # Pre-evaluation
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Add to common Y axis", position=position)
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

        with self.subTest("2Channels"):
            # Setup
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            position_c = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()

            # Pre-evaluation
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            # Event

            self.context_menu(action_text="Remove from common Y axis", position=position_c)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Add to common Y axis", position=position_c)
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

    def test_Action_SetColor(self):
        """
        Test Scope:
            - Ensure that channel color is changed.
        Events:
            - Open Context Menu
            - Select 'Set color [C]'
            - Select 1 Channel
            - Open Context Menu
            - Select 'Set color [C]'
            - Select 2 Channels
            - Open Context Menu
            - Select 'Set color [C]'
        Evaluate:
            - Evaluate that color dialog is not open if channel is not selected.
            - Evaluate that channel color is changed.
        """
        action_text = "Set color [C]"

        # Event
        with self.subTest("NoChannelSelected"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                self.context_menu(action_text=action_text)
                mo_getColor.assert_not_called()

        with self.subTest("1ChannelSelected"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                # Setup
                previous_color = self.plot_channel_a.color.name()
                color = QtGui.QColor("red")
                color_name = color.name()
                mo_getColor.return_value = color
                # Event
                self.context_menu(action_text=action_text, position=position)
                # Evaluate
                current_color = self.plot_channel_a.color.name()
                mo_getColor.assert_called()
                self.assertNotEqual(previous_color, current_color)
                self.assertEqual(color_name, current_color)

        with self.subTest("2ChannelsSelected"):
            self.mouseClick_WidgetItem(self.plot_channel_b)
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                # Setup
                previous_b_color = self.plot_channel_b.color.name()
                previous_c_color = self.plot_channel_c.color.name()
                color = QtGui.QColor("blue")
                color_name = color.name()
                mo_getColor.return_value = color
                # Event
                self.context_menu(action_text=action_text, position=position)
                # Evaluate
                current_b_color = self.plot_channel_b.color.name()
                current_c_color = self.plot_channel_c.color.name()
                mo_getColor.assert_called()
                self.assertNotEqual(previous_b_color, current_b_color)
                self.assertNotEqual(previous_c_color, current_c_color)
                self.assertEqual(color_name, current_b_color)
                self.assertEqual(color_name, current_c_color)

    def test_Action_SetRandomColor(self):
        """
        Test Scope:
            - Ensure that channel color is changed.
        Events:
            - Open Context Menu
            - Select 'Set random color'
            - Select 1 Channel
            - Open Context Menu
            - Select 'Set random color'
            - Select 2 Channels
            - Open Context Menu
            - Select 'Set random color'
        Evaluate:
            - Evaluate that color dialog is not open if channel is not selected.
            - Evaluate that channel color is changed.
        """
        action_text = "Set random color"

        # Event
        with self.subTest("NoChannelSelected"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                self.context_menu(action_text=action_text)
                mo_getColor.assert_not_called()

        with self.subTest("1ChannelSelected"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            # Setup
            previous_color = self.plot_channel_a.color.name()
            # Event
            self.context_menu(action_text=action_text, position=position)
            # Evaluate
            current_color = self.plot_channel_a.color.name()
            self.assertNotEqual(previous_color, current_color)

        with self.subTest("2ChannelsSelected"):
            self.mouseClick_WidgetItem(self.plot_channel_b)
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            # Setup
            previous_b_color = self.plot_channel_b.color.name()
            previous_c_color = self.plot_channel_c.color.name()
            # Event
            self.context_menu(action_text=action_text, position=position)
            # Evaluate
            current_b_color = self.plot_channel_b.color.name()
            current_c_color = self.plot_channel_c.color.name()
            self.assertNotEqual(previous_b_color, current_b_color)
            self.assertNotEqual(previous_c_color, current_c_color)
            self.assertNotEqual(current_b_color, current_c_color)

    @unittest.skipIf(sys.platform == "win32", "times out on Windows")
    def test_Action_CopyDisplayProperties_Group(self):
        """
        Test Scope:
            - Ensure that first channel display properties are copied in clipboard if item selected is group item.
        Events:
            - Insert Empty Group
            - Select Group
            - Trigger 'Copy display properties' action
            - Add items to group
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Remove channel from group
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that channel display properties are stored in clipboard in json format.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")
            mo_getText.return_value = "B", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")
            mo_getText.return_value = "C", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        group_channel_a = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        group_channel_b = self.plot.channel_selection.findItems("B", QtCore.Qt.MatchFlags())[0]
        group_channel_c = self.plot.channel_selection.findItems("C", QtCore.Qt.MatchFlags())[0]

        QtWidgets.QApplication.instance().clipboard().clear()
        with self.subTest("EmptyGroup_0"):
            position = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text="Copy display properties [Ctrl+Shift+C]", position=position)
            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "group")

        with self.subTest("PopulatedGroup"):
            # Add Channels to Group
            self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel_a)

            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Copy display properties [Ctrl+Shift+C]", position=position)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "channel")

        with self.subTest("EmptyGroup_1"):
            group_channel_a.removeChild(self.plot_channel_a)
            position = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text="Copy display properties [Ctrl+Shift+C]", position=position)
            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "group")

        with self.subTest("EmptyGroup_2"):
            # Add Channels to Group
            self.move_item_inside_channels_tree_widget(src=group_channel_c, dst=group_channel_b)
            group_channel_a.setExpanded(True)
            group_channel_b.setExpanded(True)
            group_channel_c.setExpanded(True)
            self.move_item_inside_channels_tree_widget(src=group_channel_b, dst=group_channel_a)
            group_channel_a.setExpanded(True)
            group_channel_b.setExpanded(True)
            group_channel_c.setExpanded(True)
            position = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text="Copy display properties [Ctrl+Shift+C]", position=position)
            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "group")

    def test_Action_PasteDisplayProperties_Group(self):
        """
        Test Scope:
            - Ensure that channel display properties can be pasted over different channel.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Select a group channel
            - Open Context Menu
            - Trigger 'Paste display properties' action
            - Select group channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Select a channel
            - Open Context Menu
            - Trigger 'Paste display properties' action
        Evaluate:
            - Evaluate that display properties are transferred from one channel to another
        """
        action_copy = "Copy display properties [Ctrl+Shift+C]"
        action_paste = "Paste display properties [Ctrl+Shift+V]"

        # Insert Group
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")
            mo_getText.return_value = "B", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")
            mo_getText.return_value = "C", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel_a = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        group_channel_a.setExpanded(True)
        group_channel_c = self.plot.channel_selection.findItems("C", QtCore.Qt.MatchFlags())[0]
        group_channel_c.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=group_channel_c, dst=group_channel_a)
        group_channel_a.setExpanded(True)
        group_channel_c.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_b, dst=group_channel_c)

        # Add Channels to Group
        group_channel_b = self.plot.channel_selection.findItems("B", QtCore.Qt.MatchFlags())[0]
        group_channel_b.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_d, dst=group_channel_b)

        with self.subTest("FromChannel_ToGroup"):
            position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            channel_a_properties = QtWidgets.QApplication.instance().clipboard().text()
            channel_a_properties = json.loads(channel_a_properties)

            # Paste
            position_dst = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_paste, position=position_dst)

            # Copy
            position_src = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            group_channel_properties = QtWidgets.QApplication.instance().clipboard().text()
            group_channel_properties = json.loads(group_channel_properties)

            # Evaluate
            self.assertEqual(channel_a_properties["ranges"], group_channel_properties["ranges"])

        with self.subTest("FromGroup_ToChannel"):
            position_src = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            # Paste
            position_dst = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            self.context_menu(action_text=action_paste, position=position_dst)

            # Copy
            position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            self.context_menu(action_text=action_copy, position=position_src)

            channel_c_properties = QtWidgets.QApplication.instance().clipboard().text()
            channel_c_properties = json.loads(channel_c_properties)

            # Evaluate
            self.assertEqual(channel_c_properties["ranges"], group_channel_properties["ranges"])

        with self.subTest("FromGroup_ToGroup"):
            self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel_a)

            position_src = self.plot.channel_selection.visualItemRect(group_channel_b).center()
            self.context_menu(action_text=action_copy, position=position_src)
            group_channel_b_properties = QtWidgets.QApplication.instance().clipboard().text()
            group_channel_b_properties = json.loads(group_channel_b_properties)

            # Paste
            position_dst = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_paste, position=position_dst)

            # Copy
            position_src = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            group_channel_a_properties = QtWidgets.QApplication.instance().clipboard().text()
            group_channel_a_properties = json.loads(group_channel_a_properties)

            # Evaluate
            self.assertEqual(group_channel_a_properties, group_channel_b_properties)

    @unittest.skipIf(sys.platform == "win32", "times out on Windows")
    def test_Action_CopyChannelStructure_Group(self):
        """
        Test Scope:
            - Ensure that channel structure is copied in clipboard.
        Events:
            - Select one group
            - Open Context Menu
            - Trigger 'Copy channel structure' action
        Evaluate:
            - Evaluate that group channel structure is stored in clipboard in json format.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("FirstGroup", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel)

        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Copy channel structure [Ctrl+C]", position=position)

        clipboard = QtWidgets.QApplication.instance().clipboard().text()
        try:
            content = json.loads(clipboard)
        except JSONDecodeError:
            self.fail("Clipboard Content cannot be decoded as JSON content.")
        else:
            self.assertIsInstance(content, list)
            for channel_properties in content:
                self.assertIsInstance(channel_properties, dict)

    def test_Action_PasteChannelStructure_Group(self):
        """
        Test Scope:
            - Ensure that channel and structure is duplicated and structure is kept.
        Events:
            - Create group
            - Add channels to group
            - Open Context Menu
            - Trigger 'Copy channel structure' action
            - Trigger 'Paste channel structure' action over new group
        Evaluate:
            - Evaluate that group channel is duplicated and structure is kept.
        """
        action_copy = "Copy channel structure [Ctrl+C]"
        action_paste = "Paste channel structure [Ctrl+V]"

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("FirstGroup", QtCore.Qt.MatchFlags())[0]
        group_channel.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_channel)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_d, dst=group_channel)

        # Copy Channel Structure
        position_src = self.plot.channel_selection.visualItemRect(group_channel).center()
        self.context_menu(action_text=action_copy, position=position_src)

        channels_count = self.plot.channel_selection.topLevelItemCount()

        # Paste Channel Structure
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text=action_paste, position=position_src)
        self.processEvents(0.1)

        self.assertEqual(channels_count + 1, self.plot.channel_selection.topLevelItemCount())
        self.assertEqual(2, len(self.plot.channel_selection.findItems("FirstGroup", QtCore.Qt.MatchFlags())))

    def test_Menu_EnableDisable_Action_DisableSelected(self):
        """
        Test Scope:
            - Ensure that channel selected is disabled.
        Events:
            - Disable all
            - Open Context Menu
            - Select action "Disable selected" from sub-menu "Enable/disable"
            - Select two channel items
            - Open Context Menu
            - Select action "Disable selected" from sub-menu "Enable/disable"
            - Select channel group
            - Open Context Menu
            - Select action "Disable selected" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that selected items were disabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_channel)

        self.context_menu(action_text="Disable selected")

        # Ensure that all items are disabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))

        # Select two channels
        self.plot_channel_a.setSelected(True)
        self.plot_channel_b.setSelected(True)
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text="Disable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Unchecked, self.plot_channel_a.checkState(self.Column.NAME))
        self.assertEqual(QtCore.Qt.CheckState.Unchecked, self.plot_channel_b.checkState(self.Column.NAME))

        # Select group
        position_src = self.plot.channel_selection.visualItemRect(group_channel).center()
        self.context_menu(action_text="Disable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Unchecked, group_channel.checkState(self.Column.NAME))

    def test_Menu_EnableDisable_Action_DeactivateGroups_Group(self):
        """
        Test Scope:
            - Ensure that group items can be disabled from Context Menu.
        Events:
            - Add two groups
            - Add channels to group
            - Add group in group
            - Select 1 group
            - Open Context Menu
            - Trigger "Enable/Disable -> Deactivate Groups" action
        Expected:
            - Group Items should be disabled
            - Sub-groups should be disabled
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "Group_A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")
            mo_getText.return_value = "Group_B", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_a_channel = self.plot.channel_selection.findItems("Group_A", QtCore.Qt.MatchFlags())[0]
        group_a_channel.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_a_channel)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_b, dst=group_a_channel)

        # Add Channels to Group
        group_b_channel = self.plot.channel_selection.findItems("Group_B", QtCore.Qt.MatchFlags())[0]
        group_b_channel.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_b_channel)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_d, dst=group_b_channel)

        self.move_item_inside_channels_tree_widget(src=group_b_channel, dst=group_a_channel)
        group_a_channel.setExpanded(True)
        group_b_channel.setExpanded(True)

        position = self.plot.channel_selection.visualItemRect(group_b_channel).center()
        self.context_menu(action_text="Deactivate groups", position=position)

        # Evaluate
        self.assertEqual(False, group_a_channel.isDisabled())
        self.assertEqual(True, group_b_channel.isDisabled())
        for child_index in range(group_b_channel.childCount()):
            child = group_b_channel.child(child_index)
            self.assertEqual(True, child.isDisabled())
        for child_index in range(group_a_channel.childCount()):
            child = group_a_channel.child(child_index)
            if child.type() == child.Channel:
                self.assertEqual(False, child.isDisabled())

        position = self.plot.channel_selection.visualItemRect(group_a_channel).center()
        self.context_menu(action_text="Deactivate groups", position=position)

        # Evaluate
        self.assertEqual(True, group_a_channel.isDisabled())
        self.assertEqual(True, group_b_channel.isDisabled())
        for child_index in range(group_b_channel.childCount()):
            child = group_b_channel.child(child_index)
            self.assertEqual(True, child.isDisabled())
        for child_index in range(group_a_channel.childCount()):
            child = group_a_channel.child(child_index)
            self.assertEqual(True, child.isDisabled())

    def test_Menu_EnableDisable_Action_DisableAll(self):
        """
        Test Scope:
            - Ensure that all channels are disabled.
        Events:
            - Open Context Menu
            - Select action "Disable All" from sub-menu "Enable/disable"
            - Disable elements
            - Open Context Menu
            - Select action "Disable All" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that all items are disabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel)
        self.processEvents(1)

        # Ensure that all items are enabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))

        self.context_menu(action_text="Disable all")

        # Ensure that all items are disabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))
                # Disable items
                item.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)

        self.context_menu(action_text="Disable all")

        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))

    def test_Menu_EnableDisable_Action_EnableSelected(self):
        """
        Test Scope:
            - Ensure that channel selected is enabled.
        Events:
            - Disable all
            - Open Context Menu
            - Select action "Enable selected" from sub-menu "Enable/disable"
            - Select two channel items
            - Open Context Menu
            - Select action "Enable selected" from sub-menu "Enable/disable"
            - Select channel group
            - Open Context Menu
            - Select action "Enable selected" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that selected items were enabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_channel)

        # Disable all items
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                item.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Unchecked)

        self.context_menu(action_text="Enable selected")

        # Ensure that all items are disabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))

        # Select two channels
        self.plot_channel_a.setSelected(True)
        self.plot_channel_b.setSelected(True)
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text="Enable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.NAME))
        self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.NAME))

        # Select group
        position_src = self.plot.channel_selection.visualItemRect(group_channel).center()
        self.context_menu(action_text="Enable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Checked, group_channel.checkState(self.Column.NAME))

    def test_Menu_EnableDisable_Action_EnableAll(self):
        """
        Test Scope:
            - Ensure that all channels are enabled.
        Events:
            - Open Context Menu
            - Select action "Enable All" from sub-menu "Enable/disable"
            - Disable elements
            - Open Context Menu
            - Select action "Enable All" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that all items are enabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group [Shift+Insert]")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel)

        # Ensure that all items are enabled:
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))

        self.context_menu(action_text="Enable all")

        # Ensure that all items are enabled:
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))
                # Disable items
                item.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Unchecked)

        self.context_menu(action_text="Enable all")

        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))



"""asammdf utility functions for channel conversions"""

from copy import deepcopy
import typing
from typing import Literal

from typing_extensions import overload

from . import v2_v3_blocks as v3b
from . import v2_v3_constants as v3c
from . import v4_blocks as v4b
from . import v4_constants as v4c
from .types import ChannelConversionType


@overload
def conversion_transfer(
    conversion: ChannelConversionType | None, version: Literal[2, 3] = ..., copy: bool = ...
) -> v3b.ChannelConversion: ...


@overload
def conversion_transfer(
    conversion: ChannelConversionType, version: Literal[4], copy: bool = ...
) -> v4b.ChannelConversion: ...


@overload
def conversion_transfer(conversion: None, version: Literal[4], copy: bool = ...) -> None: ...


@overload
def conversion_transfer(
    conversion: ChannelConversionType | None, version: Literal[2, 3, 4] = ..., copy: bool = ...
) -> ChannelConversionType | None: ...


def conversion_transfer(
    conversion: ChannelConversionType | None, version: Literal[2, 3, 4] = 3, copy: bool = False
) -> ChannelConversionType | None:
    """Convert between MDF4 and MDF3 channel conversions.

    Parameters
    ----------
    conversion : block
        Channel conversion.
    version : int, default 3
        Target MDF version.
    copy : bool, default False
        Return a copy if the input conversion version is the same as the
        required version.

    Returns
    -------
    conversion : block
        Channel conversion for specified version.
    """

    if version <= 3:
        if conversion is None:
            conversion = v3b.ChannelConversion(conversion_type=v3c.CONVERSION_TYPE_NONE)
        else:
            conversion_type = conversion.conversion_type
            if isinstance(conversion, v3b.ChannelConversion):
                if copy:
                    conversion = deepcopy(conversion)
            else:
                unit = conversion.unit.strip(" \r\n\t\0").encode("latin-1")

                match conversion_type:
                    case v4c.CONVERSION_TYPE_NON:
                        conversion = v3b.ChannelConversion(unit=unit, conversion_type=v3c.CONVERSION_TYPE_NONE)

                    case v4c.CONVERSION_TYPE_LIN:
                        conversion = v3b.ChannelConversion(
                            unit=unit,
                            conversion_type=v3c.CONVERSION_TYPE_LINEAR,
                            a=conversion.a,
                            b=conversion.b,
                        )

                    case v4c.CONVERSION_TYPE_RAT:
                        conversion = v3b.ChannelConversion(
                            unit=unit,
                            conversion_type=v3c.CONVERSION_TYPE_RAT,
                            P1=conversion.P1,
                            P2=conversion.P2,
                            P3=conversion.P3,
                            P4=conversion.P4,
                            P5=conversion.P5,
                            P6=conversion.P6,
                        )

                    case v4c.CONVERSION_TYPE_TAB:
                        v3_kwargs = v3b.ChannelConversionKwargs(unit=unit, conversion_type=v3c.CONVERSION_TYPE_TAB)
                        v3_kwargs["ref_param_nr"] = conversion.val_param_nr // 2
                        for i in range(conversion.val_param_nr // 2):
                            v3_kwargs[f"raw_{i}"] = conversion[f"raw_{i}"]  # type: ignore[literal-required]
                            v3_kwargs[f"phys_{i}"] = conversion[f"phys_{i}"]  # type: ignore[literal-required]

                        conversion = v3b.ChannelConversion(**v3_kwargs)

                    case v4c.CONVERSION_TYPE_TABI:
                        v3_kwargs = v3b.ChannelConversionKwargs(unit=unit, conversion_type=v3c.CONVERSION_TYPE_TABI)
                        v3_kwargs["ref_param_nr"] = conversion.val_param_nr // 2
                        for i in range(conversion.val_param_nr // 2):
                            v3_kwargs[f"raw_{i}"] = conversion[f"raw_{i}"]  # type: ignore[literal-required]
                            v3_kwargs[f"phys_{i}"] = conversion[f"phys_{i}"]  # type: ignore[literal-required]

                        conversion = v3b.ChannelConversion(**v3_kwargs)

                    case v4c.CONVERSION_TYPE_ALG:
                        formula = conversion.formula.replace("X", "X1")
                        conversion = v3b.ChannelConversion(
                            formula=formula,
                            unit=unit,
                            conversion_type=v3c.CONVERSION_TYPE_FORMULA,
                        )

                    case v4c.CONVERSION_TYPE_RTAB:
                        nr = (conversion.val_param_nr - 1) // 3
                        v3_kwargs = {
                            "unit": unit,
                            "ref_param_nr": nr,
                            "conversion_type": v3c.CONVERSION_TYPE_TABI,
                        }

                        for i in range(nr):
                            l_ = conversion[f"lower_{i}"]
                            u_ = typing.cast(float, conversion[f"upper_{i}"])
                            p_ = conversion[f"phys_{i}"]
                            v3_kwargs[f"raw_{i}"] = l_  # type: ignore[literal-required]
                            v3_kwargs[f"raw_{i}"] = u_ - 0.000_001  # type: ignore[literal-required]
                            v3_kwargs[f"phys_{i}"] = p_  # type: ignore[literal-required]
                            v3_kwargs[f"phys_{i}"] = p_  # type: ignore[literal-required]

                        conversion = v3b.ChannelConversion(**v3_kwargs)

                    case v4c.CONVERSION_TYPE_TABX:
                        nr = conversion.val_param_nr

                        v3_kwargs = {
                            "ref_param_nr": nr + 1,
                            "unit": unit,
                            "conversion_type": v3c.CONVERSION_TYPE_RTABX,
                        }
                        for i in range(nr):
                            v3_kwargs[f"lower_{i}"] = conversion[f"val_{i}"]  # type: ignore[literal-required]
                            v3_kwargs[f"upper_{i}"] = conversion[f"val_{i}"]  # type: ignore[literal-required]
                            block = conversion.referenced_blocks[f"text_{i}"]
                            if isinstance(block, v4b.ChannelConversion):
                                v3_kwargs[f"text_{i}"] = block.name.encode("latin-1")  # type: ignore[literal-required]
                            else:
                                v3_kwargs[f"text_{i}"] = block  # type: ignore[literal-required]

                        new_conversion = v3b.ChannelConversion(**v3_kwargs)
                        if isinstance(
                            conversion.referenced_blocks["default_addr"],
                            v4b.ChannelConversion,
                        ):
                            default_addr = conversion.referenced_blocks["default_addr"].name.encode("latin-1")
                        else:
                            default_addr = conversion.referenced_blocks["default_addr"]
                        new_conversion.referenced_blocks["default_addr"] = default_addr

                        conversion = new_conversion

                    case v4c.CONVERSION_TYPE_RTABX:
                        nr = conversion.val_param_nr // 2
                        v3_kwargs = {
                            "ref_param_nr": nr + 1,
                            "unit": unit,
                            "conversion_type": v3c.CONVERSION_TYPE_RTABX,
                        }
                        for i in range(nr):
                            v3_kwargs[f"lower_{i}"] = conversion[f"lower_{i}"]  # type: ignore[literal-required]
                            v3_kwargs[f"upper_{i}"] = conversion[f"upper_{i}"]  # type: ignore[literal-required]
                            block = conversion.referenced_blocks[f"text_{i}"]
                            if isinstance(block, v4b.ChannelConversion):
                                v3_kwargs[f"text_{i}"] = block.name.encode("latin-1")  # type: ignore[literal-required]
                            else:
                                v3_kwargs[f"text_{i}"] = block  # type: ignore[literal-required]

                        new_conversion = v3b.ChannelConversion(**v3_kwargs)
                        if isinstance(
                            conversion.referenced_blocks["default_addr"],
                            v4b.ChannelConversion,
                        ):
                            default_addr = conversion.referenced_blocks["default_addr"].name.encode("latin-1")
                        else:
                            default_addr = conversion.referenced_blocks["default_addr"]
                        new_conversion.referenced_blocks["default_addr"] = default_addr

                        conversion = new_conversion

    else:
        if not conversion or isinstance(conversion, v4b.ChannelConversion):
            if copy:
                conversion = deepcopy(conversion)
        else:
            conversion_type = conversion.conversion_type
            unit_str = conversion.unit_field.decode("latin-1").strip(" \r\n\t\0")

            match conversion_type:
                case v3c.CONVERSION_TYPE_NONE:
                    conversion = v4b.ChannelConversion(conversion_type=v4c.CONVERSION_TYPE_NON)

                case v3c.CONVERSION_TYPE_LINEAR:
                    conversion = v4b.ChannelConversion(
                        conversion_type=v4c.CONVERSION_TYPE_LIN,
                        a=conversion.a,
                        b=conversion.b,
                    )

                case v3c.CONVERSION_TYPE_RAT:
                    conversion = v4b.ChannelConversion(
                        conversion_type=v4c.CONVERSION_TYPE_RAT,
                        P1=conversion.P1,
                        P2=conversion.P2,
                        P3=conversion.P3,
                        P4=conversion.P4,
                        P5=conversion.P5,
                        P6=conversion.P6,
                    )

                case v3c.CONVERSION_TYPE_FORMULA:
                    formula = conversion.formula
                    conversion = v4b.ChannelConversion(conversion_type=v4c.CONVERSION_TYPE_ALG, formula=formula)

                case v3c.CONVERSION_TYPE_TAB:
                    v4_kwargs = v4b.ChannelConversionKwargs(conversion_type=v4c.CONVERSION_TYPE_TAB)
                    v4_kwargs["val_param_nr"] = conversion.ref_param_nr * 2
                    for i in range(conversion.ref_param_nr):
                        v4_kwargs[f"raw_{i}"] = conversion[f"raw_{i}"]  # type: ignore[literal-required]
                        v4_kwargs[f"phys_{i}"] = conversion[f"phys_{i}"]  # type: ignore[literal-required]

                    conversion = v4b.ChannelConversion(**v4_kwargs)

                case v3c.CONVERSION_TYPE_TABI:
                    v4_kwargs = v4b.ChannelConversionKwargs(conversion_type=v4c.CONVERSION_TYPE_TABI)
                    v4_kwargs["val_param_nr"] = conversion.ref_param_nr * 2
                    for i in range(conversion.ref_param_nr):
                        v4_kwargs[f"raw_{i}"] = conversion[f"raw_{i}"]  # type: ignore[literal-required]
                        v4_kwargs[f"phys_{i}"] = conversion[f"phys_{i}"]  # type: ignore[literal-required]

                    conversion = v4b.ChannelConversion(**v4_kwargs)

                case v3c.CONVERSION_TYPE_TABX:
                    nr = conversion.ref_param_nr
                    v4_kwargs = {
                        "val_param_nr": nr,
                        "ref_param_nr": nr + 1,
                        "conversion_type": v4c.CONVERSION_TYPE_TABX,
                    }
                    for i in range(nr):
                        v4_kwargs[f"val_{i}"] = conversion[f"param_val_{i}"]  # type: ignore[literal-required]
                        v4_kwargs[f"text_{i}"] = conversion[f"text_{i}"]  # type: ignore[literal-required]

                    conversion = v4b.ChannelConversion(**v4_kwargs)

                case v3c.CONVERSION_TYPE_RTABX:
                    nr = conversion.ref_param_nr - 1
                    v4_kwargs = {
                        "val_param_nr": nr * 2,
                        "ref_param_nr": nr + 1,
                        "conversion_type": v4c.CONVERSION_TYPE_RTABX,
                        "default_addr": typing.cast(bytes, conversion.referenced_blocks["default_addr"]),
                    }
                    for i in range(nr):
                        v4_kwargs[f"lower_{i}"] = conversion[f"lower_{i}"]  # type: ignore[literal-required]
                        v4_kwargs[f"upper_{i}"] = conversion[f"upper_{i}"]  # type: ignore[literal-required]
                        v4_kwargs[f"text_{i}"] = conversion.referenced_blocks[f"text_{i}"]  # type: ignore[literal-required]

                    conversion = v4b.ChannelConversion(**v4_kwargs)

            conversion.unit = unit_str

    return conversion


def inverse_conversion(
    conversion: (
        ChannelConversionType | v3b.ChannelConversionKwargs | v4b.ChannelConversionKwargs | dict[str, object] | None
    ),
) -> v4b.ChannelConversion | None:
    if not conversion:
        return None

    if isinstance(conversion, v3b.ChannelConversion):
        conversion = conversion_transfer(conversion, version=4)

    conversion_dict: v3b.ChannelConversionKwargs | v4b.ChannelConversionKwargs | dict[str, object]
    if not isinstance(conversion, dict):
        conversion_dict = to_dict(conversion)
    else:
        conversion_dict = conversion

    if "a" in conversion_dict:
        kwargs: v4b.ChannelConversionKwargs = {
            "conversion_type": v4c.CONVERSION_TYPE_LIN,
            "a": 1 / typing.cast(float, conversion_dict["a"]),
            "b": typing.cast(float, conversion_dict["b"]) / typing.cast(float, conversion_dict["a"]),
        }
        conv = v4b.ChannelConversion(**kwargs)

    elif "P1" in conversion_dict:
        a, b, c, d, e, f = (typing.cast(float, conversion_dict[f"P{i}"]) for i in range(1, 7))  # type: ignore[literal-required]

        if e == 0 and f == 0:
            if d == 0 and a == 0:
                conv = None
            else:
                kwargs = {
                    "P1": 0,
                    "P2": e,
                    "P3": -b,
                    "P4": 0,
                    "P5": -d,
                    "P6": a,
                    "conversion_type": v4c.CONVERSION_TYPE_RAT,
                }
                conv = v4b.ChannelConversion(**kwargs)

        elif a == 0 and d == 0:
            if e == 0 and b == 0:
                conv = None
            else:
                kwargs = {
                    "P1": 0,
                    "P2": -f,
                    "P3": c,
                    "P4": 0,
                    "P5": e,
                    "P6": -b,
                    "conversion_type": v4c.CONVERSION_TYPE_RAT,
                }
                conv = v4b.ChannelConversion(**kwargs)

    else:
        conv = None

    return conv


@overload
def from_dict(conversion_dict: v4b.ChannelConversionKwargs | dict[str, object]) -> v4b.ChannelConversion: ...


@overload
def from_dict(conversion_dict: None) -> None: ...


@overload
def from_dict(
    conversion_dict: v4b.ChannelConversionKwargs | dict[str, object] | None,
) -> v4b.ChannelConversion | None: ...


def from_dict(conversion_dict: v4b.ChannelConversionKwargs | dict[str, object] | None) -> v4b.ChannelConversion | None:
    if not conversion_dict:
        conversion = None

    elif "a" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_LIN
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "formula" in conversion_dict:
        conversion_dict["formula"] = typing.cast(str, conversion_dict["formula"])
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_ALG
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif all(key in conversion_dict for key in [f"P{i}" for i in range(1, 7)]):
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_RAT
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "raw_0" in conversion_dict and "phys_0" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_TAB
        nr = 0
        while f"phys_{nr}" in conversion_dict:
            nr += 1
        conversion_dict["val_param_nr"] = nr * 2
        if conversion_dict.get("interpolation", False):
            conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_TABI
        else:
            conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_TAB
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "mask_0" in conversion_dict and "text_0" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_BITFIELD
        nr = 0
        while f"text_{nr}" in conversion_dict:
            val = conversion_dict[f"text_{nr}"]  # type: ignore[literal-required]
            if isinstance(val, (bytes, str)):
                partial_conversion: dict[str, object] = {
                    "conversion_type": v4c.CONVERSION_TYPE_RTABX,
                    f"upper_{nr}": conversion_dict[f"upper_{nr}"],  # type: ignore[literal-required]
                    f"lower_{nr}": conversion_dict[f"lower_{nr}"],  # type: ignore[literal-required]
                    f"text_{nr}": (
                        text
                        if isinstance((text := typing.cast(bytes | str, conversion_dict[f"text_{nr}"])), bytes)  # type: ignore[literal-required]
                        else text.encode("utf-8")
                    ),
                    "default": b"",
                }
                conversion_dict[f"text_{nr}"] = from_dict(partial_conversion)  # type: ignore[literal-required]
            elif isinstance(val, dict):
                conversion_dict[f"text_{nr}"] = from_dict(val)  # type: ignore[literal-required]

            nr += 1

        conversion_dict["ref_param_nr"] = nr
        conversion_dict["val_param_nr"] = nr
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "upper_0" in conversion_dict and "phys_0" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_RTAB
        nr = 0
        while f"phys_{nr}" in conversion_dict:
            nr += 1
        conversion_dict["val_param_nr"] = nr * 3 + 1
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "val_0" in conversion_dict and "text_0" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_TABX
        nr = 0
        while f"text_{nr}" in conversion_dict:
            val = conversion_dict[f"text_{nr}"]  # type: ignore[literal-required]
            if isinstance(val, str):
                conversion_dict[f"text_{nr}"] = val.encode("utf-8")  # type: ignore[literal-required]
            elif isinstance(val, dict):
                conversion_dict[f"text_{nr}"] = from_dict(val)  # type: ignore[literal-required]
            nr += 1

        val = conversion_dict.get("default_addr", b"")
        if isinstance(val, str):
            conversion_dict["default_addr"] = val.encode("utf-8")
        elif isinstance(val, dict):
            conversion_dict["default_addr"] = from_dict(val)

        conversion_dict["ref_param_nr"] = nr + 1
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "upper_0" in conversion_dict and "text_0" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_RTABX
        nr = 0
        while f"text_{nr}" in conversion_dict:
            val = conversion_dict[f"text_{nr}"]  # type: ignore[literal-required]
            if isinstance(val, str):
                conversion_dict[f"text_{nr}"] = val.encode("utf-8")  # type: ignore[literal-required]
            elif isinstance(val, dict):
                conversion_dict[f"text_{nr}"] = from_dict(val)  # type: ignore[literal-required]
            nr += 1

        conversion_dict["ref_param_nr"] = nr + 1

        val = conversion_dict.get("default_addr", b"")
        if isinstance(val, str):
            conversion_dict["default_addr"] = val.encode("utf-8")
        elif isinstance(val, dict):
            conversion_dict["default_addr"] = from_dict(val)
        conversion = v4b.ChannelConversion(**conversion_dict)

    elif "default_addr" in conversion_dict:
        conversion_dict["conversion_type"] = v4c.CONVERSION_TYPE_TABX
        val = conversion_dict["default_addr"]
        if isinstance(val, str):
            conversion_dict["default_addr"] = val.encode("utf-8")
        elif isinstance(val, dict):
            conversion_dict["default_addr"] = from_dict(val)
        conversion_dict["ref_param_nr"] = 1
        conversion = v4b.ChannelConversion(**conversion_dict)

    else:
        conversion = v4b.ChannelConversion(conversion_type=v4c.CONVERSION_TYPE_NON)

    return conversion


@overload
def to_dict(conversion: ChannelConversionType) -> dict[str, object]: ...


@overload
def to_dict(conversion: None) -> None: ...


@overload
def to_dict(conversion: ChannelConversionType | None) -> dict[str, object] | None: ...


def to_dict(conversion: ChannelConversionType | None) -> dict[str, object] | None:
    if not conversion:
        return None

    if isinstance(conversion, v3b.ChannelConversion):
        conversion_v4 = conversion_transfer(conversion, version=4)
    else:
        conversion_v4 = conversion

    conversion_type = conversion_v4.conversion_type

    conversion_dict: dict[str, object] = {
        "name": conversion_v4.name,
        "unit": conversion_v4.unit,
        "comment": conversion_v4.comment,
    }

    match conversion_type:
        case v4c.CONVERSION_TYPE_LIN:
            conversion_dict["a"] = conversion_v4.a
            conversion_dict["b"] = conversion_v4.b
            conversion_dict["conversion_type"] = conversion_type

        case v4c.CONVERSION_TYPE_ALG:
            conversion_dict["formula"] = conversion_v4.formula
            conversion_dict["conversion_type"] = conversion_type

        case v4c.CONVERSION_TYPE_RAT:
            conversion_dict.update({key: conversion_v4[key] for key in [f"P{i}" for i in range(1, 7)]})
            conversion_dict["conversion_type"] = conversion_type

        case v4c.CONVERSION_TYPE_TAB | v4c.CONVERSION_TYPE_TABI:
            params = conversion_v4.val_param_nr // 2
            conversion_dict.update({key: conversion_v4[key] for key in [f"phys_{nr}" for nr in range(params)]})
            conversion_dict.update({key: conversion_v4[key] for key in [f"raw_{nr}" for nr in range(params)]})
            conversion_dict["conversion_type"] = conversion_type

        case v4c.CONVERSION_TYPE_RTAB:
            params = (conversion_v4.val_param_nr - 1) // 3
            conversion_dict.update({key: conversion_v4[key] for key in [f"lower_{nr}" for nr in range(params)]})
            conversion_dict.update({key: conversion_v4[key] for key in [f"upper_{nr}" for nr in range(params)]})
            conversion_dict.update({key: conversion_v4[key] for key in [f"phys_{nr}" for nr in range(params)]})
            conversion_dict["conversion_type"] = conversion_type
            conversion_dict["default"] = conversion_v4.default

        case v4c.CONVERSION_TYPE_TABX:
            nr = conversion_v4.ref_param_nr - 1

            conversion_dict["conversion_type"] = conversion_type

            for key, val in conversion_v4.referenced_blocks.items():
                if isinstance(val, str):
                    conversion_dict[key] = val
                elif isinstance(val, v4b.ChannelConversion):
                    conversion_dict[key] = to_dict(val)
                elif val is None:
                    conversion_dict[key] = ""
                else:
                    conversion_dict[key] = val.decode("utf-8", errors="replace")

            for i in range(nr):
                conversion_dict[f"val_{i}"] = conversion_v4[f"val_{i}"]

        case v4c.CONVERSION_TYPE_RTABX:
            nr = conversion_v4.ref_param_nr - 1

            conversion_dict["conversion_type"] = conversion_type

            for key, val in conversion_v4.referenced_blocks.items():
                if isinstance(val, str):
                    conversion_dict[key] = val
                elif isinstance(val, v4b.ChannelConversion):
                    conversion_dict[key] = to_dict(val)
                elif val is None:
                    conversion_dict[key] = ""
                else:
                    conversion_dict[key] = val.decode("utf-8", errors="replace")

            for i in range(nr):
                conversion_dict[f"upper_{i}"] = conversion_v4[f"upper_{i}"]
                conversion_dict[f"lower_{i}"] = conversion_v4[f"lower_{i}"]

        case _:
            return None

    return conversion_dict



#!/usr/bin/env python\

import os
import sys
import unittest
from unittest import mock

from PySide6.QtCore import QPoint, QRect, QSettings, Qt
from PySide6.QtGui import QGuiApplication, QKeySequence
from PySide6.QtTest import QTest
from PySide6.QtWidgets import QTreeWidgetItemIterator

from asammdf import mdf
from test.asammdf.gui.test_base import Pixmap
from test.asammdf.gui.widgets.test_BasePlotWidget import TestPlotWidget


class TestPlotGraphicsShortcuts(TestPlotWidget):
    def setUp(self):
        """
        Events:
            Open test measurement file
            Set sort method: Natural sort
            Create new plot widget with following setup:
                - Without dots
                - Without grid
                - With hidden bookmarks
                - Cursor without circle
        Evaluate:
            - Evaluate that one widget was created
            - Evaluate that plot is black
        """
        super().setUp()
        settings = QSettings()
        settings.setValue("zoom_x_center_on_cursor", True)
        settings.setValue("plot_cursor_precision", 6)
        # Open measurement file
        self.setUpFileWidget(measurement_file=self.measurement_file, default=True)
        # Switch ComboBox to "Natural sort"
        self.widget.channel_view.setCurrentText("Natural sort")
        # Select channels -> Press PushButton "Create Window" -> "Plot"
        self.create_window(window_type="Plot")
        # Evaluate
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)
        # Plot object
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        # Preset for plot
        # Remove dots
        if self.plot.plot.with_dots:
            self.plot.plot.set_dots(False)
        # check if grid is available -> hide grid
        if not self.plot.hide_axes_btn.isFlat():
            QTest.mouseClick(self.plot.hide_axes_btn, Qt.MouseButton.LeftButton)
        # hide bookmarks if it's available
        if self.plot.show_bookmarks:
            self.plot.toggle_bookmarks(hide=True)
        # Ensure that plot is not` in Focused mode
        if self.plot.focused_mode:
            self.plot.toggle_focused_mode()
        # Ensure that plot cursor is not with circle
        if self.plot.plot.cursor1.show_circle:
            self.plot.plot.cursor1.show_circle = False

        self.pg = self.plot.plot  # PlotGraphics object

        self.processEvents(0.01)
        # Evaluate that plot is black
        self.assertTrue(Pixmap.is_black(self.pg.grab()))

        # get shortcuts
        self.assertIsNotNone(self.load_shortcuts_from_json_file(self.pg))
        self.processEvents()

    def test_lock_unlock_range_shortcut(self):
        """
        Test Scope:
            Check if Range Selection cursor is locked/unlocked after pressing key Y.
        Events:
            - Set default cursor color, without horizontal line and circle, with 1px line width
            - Press Key Y for range selection
            - Move Cursors
            - Press Key R for range selection
        Evaluate:
            - Evaluate that at start is only one cursor available
            - Evaluate that two cursors are available after key Y was pressed
            - Evaluate that new rectangle with different color is present
            - Evaluate that sum of rectangle areas is same with the one when plot is full black.
            - Evaluate that range selection disappear.
        """
        # Setup for cursor
        self.widget.set_cursor_options(False, False, 1, Pixmap.COLOR_CURSOR)

        # Save PixMap of clean plot
        clean_pixmap = self.pg.grab()
        # Evaluate that plot is black
        self.assertTrue(Pixmap.is_black(clean_pixmap))

        # Get position of Cursor
        cursors = Pixmap.cursors_x(clean_pixmap)
        # Evaluate that there is only one cursor
        self.assertEqual(1, len(cursors))

        # Press Key 'Y' for range selection
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["toggle_range"]))
        self.processEvents(timeout=0.1)

        # Save PixMap of Range plot
        range_pixmap = self.pg.grab()
        self.assertFalse(Pixmap.is_black(range_pixmap))

        # Get X position of Cursors
        cursors = Pixmap.cursors_x(range_pixmap)
        # Evaluate that two cursors are available
        self.assertEqual(2, len(cursors))

        # Evaluate that new rectangle with different color is present
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=0,
                y=0,
                width=min(cursors) - 1,
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_RANGE,
                x=min(cursors) + 1,
                y=0,
                width=max(cursors),
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=max(cursors) + 1,
                y=0,
            )
        )

        # Move Cursors
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_left_1x"]))
        self.processEvents(timeout=0.01)
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_left_20x"]))
        self.processEvents(timeout=0.01)

        # Save PixMap of Range plot
        range_pixmap = self.pg.grab()
        self.assertFalse(Pixmap.is_black(range_pixmap))

        # Get X position of Cursors
        new_cursors = Pixmap.cursors_x(range_pixmap)
        # Evaluate that two cursors are available
        self.assertEqual(2, len(new_cursors))
        self.assertEqual(cursors[0], new_cursors[0], "First cursor have new position after manipulation")
        self.assertNotEqual(cursors[1], new_cursors[1], "Second cursors have same position after manipulation")

        # Evaluate that new rectangle with different color is present
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=0,
                y=0,
                width=min(new_cursors) - 1,
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_RANGE,
                x=min(new_cursors) + 1,
                y=0,
                width=max(new_cursors),
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=max(new_cursors) + 1,
                y=0,
            )
        )

        cursors = new_cursors
        # Press Key 'R' for range selection
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["toggle_range"]))
        self.processEvents(timeout=0.01)

        # Move Cursors
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_right_1x"]))
        self.processEvents(timeout=0.01)
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_left_20x"]))
        self.processEvents(timeout=0.01)

        # Save PixMap of Range plot
        range_pixmap = self.pg.grab()
        self.assertFalse(Pixmap.is_black(range_pixmap))

        # Get X position of Cursors
        new_cursors = Pixmap.cursors_x(range_pixmap)
        # Evaluate that two cursors are available
        self.assertEqual(2, len(new_cursors))
        for c in cursors:
            self.assertNotIn(c, new_cursors, f"cursor {c} is the same")

        # Evaluate that new rectangle with different color is present
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=0,
                y=0,
                width=min(new_cursors) - 1,
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_RANGE,
                x=min(new_cursors) + 1,
                y=0,
                width=max(new_cursors),
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=max(new_cursors) + 1,
                y=0,
            )
        )

        # Press Key 'R' for range selection
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["range"]))
        self.processEvents(timeout=0.01)
        # Save PixMap of clear plot
        clean_pixmap = self.pg.grab()
        self.assertTrue(Pixmap.is_black(clean_pixmap))

    def test_zoom_to_range_shortcut(self):
        """
        Test Scope:
            Check if fitting between cursors is released after pressing key "X".
        Events:
            - Display 1 signal on plot
            - Maximize window
            - Mouse click in the middle of plot
            - Press "R"
            - Press "X"
        Evaluate:
            - Evaluate that intersection of signal and midd line is exactly how much it intersects between cursors
            - Evaluate X range, it must be almost equal as range value after pressing key R
        """
        # Setup
        self.add_channels([35])
        channel_color = self.channels[0].color.name()

        # Count intersections between middle line and signal
        initial_intersections = Pixmap.color_map(self.pg.grab(QRect(0, int(self.pg.height() / 2), self.pg.width(), 1)))[
            0
        ].count(channel_color)
        self.assertTrue(initial_intersections)

        # Setup for cursor
        self.widget.set_cursor_options(False, False, 1, Pixmap.COLOR_CURSOR)
        # Mouse click on a center of plot
        QTest.mouseClick(
            self.pg.viewport(),
            Qt.MouseButton.LeftButton,
            Qt.KeyboardModifier.NoModifier,
            self.pg.rect().center(),
        )
        # Press R
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["range"]))
        self.processEvents()

        x_range = self.pg.region.getRegion()
        self.assertNotIn(x_range[0], self.pg.x_range)
        self.assertNotIn(x_range[1], self.pg.x_range)
        # Get X position of Cursor
        cursors = Pixmap.cursors_x(self.pg.grab())
        # Ensure that both cursors were found
        self.assertEqual(len(cursors), 2)

        # Get a set of colors founded between cursors
        colors = Pixmap.color_names_exclude_defaults(
            self.pg.grab(QRect(cursors[0], 0, cursors[1] - cursors[0], self.pg.height()))
        )
        # Exclude channel original color
        if channel_color in colors:
            colors.remove(channel_color)
        # caught ya
        color = colors.pop()
        # Evaluate if color was found, in set must remain only the new channel color situated between cursors
        self.assertTrue(color)

        # Count intersection of midd line and signal between cursors
        expected_intersections = Pixmap.color_map(
            self.pg.grab(QRect(cursors[0], int(self.pg.height() / 2), cursors[1] - cursors[0], 1))
        )[0].count(color)
        self.assertTrue(expected_intersections)

        # Press key "X"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["zoom_to_range"]))
        self.processEvents()

        # Evaluate how much times signal intersect midd line
        actual_intersections = Pixmap.color_map(self.pg.grab(QRect(0, int(self.pg.height() / 2), self.pg.width(), 1)))[
            0
        ].count(channel_color)
        self.assertEqual(actual_intersections, expected_intersections)
        self.assertLess(actual_intersections, initial_intersections)

        # Evaluate ranges of signal
        self.assertAlmostEqual(self.pg.x_range[0], x_range[0], delta=0.001)
        self.assertAlmostEqual(self.pg.x_range[1], x_range[1], delta=0.001)

    # @unittest.skipIf(sys.platform != "win32", "Unpredicted linux failed test")
    def test_fit__stack_shortcuts(self):
        """
        Test Scope:
            Check if:
              > all signals is stack after pressing key "S"
              > only selected signal is fitted after pressing combination "Sift + F"
              > only selected signal is stacked after pressing combination "Shift + S"
              > all signals is fitted after pressing key "F"

        Events:
            - Display 3 channels to plot
            - Press Key "S"
            - Press combination "Shift + F"
            - Press combination "Shift + S"
            - press key "F"

        Evaluate:
            - Evaluate that signals are separated in top, midd and bottom third of plot after pressing key "S"
            - Evaluate that only selected signal is fitted after pressing combination "Shift + F"
            - Evaluate that only selected signal is stacked after pressing combination "Shift + S"
            - Evaluate that all signals are fitted after pressing key "F"

        Additional Evaluation
            - Evaluate that all signals are continuous on plot
        """

        def continuous(ch):
            pixmap = self.pg.grab()
            image = pixmap.toImage()
            signal_color = ch.color.name()
            start, stop = Pixmap.search_signal_extremes_by_ax(pixmap, signal_color=signal_color, ax="x")
            for x in range(start, stop + 1):
                for j in range(self.pg.height()):
                    if image.pixelColor(x, j).name() == signal_color:
                        break

                else:
                    raise Exception(f"column {x} doesn't have color of channel {ch.name} from {start=} to {stop=}")

        self.pg.cursor1.color = "#000000"

        self.add_channels([35, 36, 37])

        channel_35 = self.channels[0]
        channel_36 = self.channels[1]
        channel_37 = self.channels[2]
        # Press "S"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["stack_all"]))
        self.processEvents()
        # Evaluate
        with self.subTest("test_stack_all_shortcut"):
            # First 2 lines
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, 0, self.pg.width(), 1))))
            # Top
            pixmap = self.pg.grab(QRect(0, 0, self.pg.width(), int(self.pg.height() / 3)))
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Midd
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3),
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertFalse(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Bottom
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3) * 2,
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertFalse(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Last 2 lines

            pixmap = self.pg.grab(QRect(0, self.pg.height() - 3, self.pg.width(), 2))
            cn = Pixmap.color_names_exclude_defaults(pixmap)
            self.assertEqual(len(cn), 0)

        # select the first channel
        self.mouseClick_WidgetItem(channel_35)
        # Press "Shift+F"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["fit_selected"]))
        self.avoid_blinking_issue(self.plot.channel_selection)
        for _ in range(50):
            self.processEvents()
        # Evaluate
        with self.subTest("test_fit_selected_shortcut"):
            # First line
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, 0, self.pg.width(), 1))))
            # Top
            pixmap = self.pg.grab(QRect(0, 0, self.pg.width(), int(self.pg.height() / 3)))
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Midd
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3),
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Bottom
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3) * 2,
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Last line
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, self.pg.height() - 2, self.pg.width(), 1))))

        # select second channel
        self.mouseClick_WidgetItem(channel_36)
        # Press "Shift+F"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["stack_selected"]))
        self.avoid_blinking_issue(self.plot.channel_selection)
        # Evaluate
        with self.subTest("test_stack_selected_shortcut"):
            # First line
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, 0, self.pg.width(), 1))))
            # Top
            pixmap = self.pg.grab(QRect(0, 0, self.pg.width(), int(self.pg.height() / 3)))
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Midd
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3),
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertFalse(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Bottom
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3) * 2,
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Last line
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, self.pg.height() - 1, self.pg.width(), 1))))

            # Press "F"
            QTest.keySequence(self.pg, QKeySequence(self.shortcuts["fit_all"]))
            self.avoid_blinking_issue(self.plot.channel_selection)
            # Evaluate
        with self.subTest("test_fit_all_shortcut"):
            # First line
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, 0, self.pg.width(), 1))))
            # Top
            pixmap = self.pg.grab(QRect(0, 0, self.pg.width(), int(self.pg.height() / 3)))
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Midd
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3),
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Bottom
            pixmap = self.pg.grab(
                QRect(
                    0,
                    int(self.pg.height() / 3) * 2,
                    self.pg.width(),
                    int(self.pg.height() / 3),
                )
            )
            self.assertTrue(Pixmap.has_color(pixmap, channel_35.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_36.color.name()))
            self.assertTrue(Pixmap.has_color(pixmap, channel_37.color.name()))
            # Last line
            self.assertTrue(Pixmap.is_black(self.pg.grab(QRect(0, self.pg.height() - 1, self.pg.width(), 1))))

            # deselect all channels
            for channel in self.channels:
                self.mouseDClick_WidgetItem(channel)

            # search if all channels are fitted into extremes
            self.mouseDClick_WidgetItem(channel_35)
            continuous(channel_35)

            self.mouseDClick_WidgetItem(channel_35)
            self.mouseDClick_WidgetItem(channel_36)
            continuous(channel_36)

            self.mouseDClick_WidgetItem(channel_36)
            self.mouseDClick_WidgetItem(channel_37)
            continuous(channel_37)

    def test_grid_shortcut(self):
        """
        Test Scope:
            Check if grid is created properly after pressing key "G".

        Events:
            - If axes are hidden - press "Show axes" button
            - Press Key "G" 20 times

        Evaluate:
            - Evaluate that grid is displayed in order after pressing key "G":
                1. Is only X axis grid.
                2. Is X and Y axes grids.
                3. Is only Y axis grid
                4. There is no grid.
        """
        # Check if the grid is available

        if self.plot.hide_axes_btn.isFlat():
            QTest.mouseClick(self.plot.hide_axes_btn, Qt.MouseButton.LeftButton)

        next_grid = {
            (False, False): (True, False),
            (True, False): (True, True),
            (True, True): (False, True),
            (False, True): (False, False),
        }

        current_grid = self.pg.x_axis.grid, self.pg.y_axis.grid
        for i in range(20):
            # press key "G"
            QTest.keySequence(self.pg, QKeySequence(self.shortcuts["grid"]))
            self.processEvents()
            # Evaluate
            current_grid = next_grid[current_grid]
            self.assertEqual(current_grid, (self.pg.x_axis.grid, self.pg.y_axis.grid))

    def test_go_to_timestamp_shortcut(self):
        """
        Test scope:
            Ensure that Shift+G will switchto selected timestamp

        Events:
            - Display one item to the widget.
            - Press Shift+G -> input some value -> "Ok".

        Evaluate:
             - Evaluate that there is added one item to the widget.
             - Evaluate that getDouble object was called.
             - Evaluate that timestamp value is almost equal with the inputted value.
        Returns
        -------

        """
        # Setup
        # Add one item to the widget
        self.assertIsNotNone(self.add_channels([10]))

        # Event
        with mock.patch("asammdf.gui.widgets.plot.QtWidgets.QInputDialog.getDouble") as mo_getDouble:
            expected_pos = 3.0
            mo_getDouble.return_value = expected_pos, True
            # Press Shift+G
            QTest.keySequence(self.plot, QKeySequence(self.shortcuts["go_to_timestamp"]))
            self.processEvents(0.1)
        self.processEvents(0.01)
        # Evaluate
        mo_getDouble.assert_called()
        ci = self.plot.cursor_info
        pos = float(ci.text().split()[2].split(ci.unit)[0])

        self.assertAlmostEqual(expected_pos, pos, delta=0.01)

    def test_zoom_in__out_shortcuts(self):
        """
        Test Scope:
            Check if zooming is released after pressing keys "I", "Shift+I", "O", "Shift+O".

        Events:
            - Display 1 signal on plot
            - Select signal, click in the middle of plot
            - Press "I".
            - Press "O".
            - Press "Shift+I".
            - Press "Shift+O".

        Evaluate:
            - Evaluate ranges of the "X" and the "Y" asis.
        """

        def get_expected_result(step, is_x_axis: bool):
            if is_x_axis:
                delta = self.pg.x_range[1] - self.pg.x_range[0]
                val = self.pg.cursor1.value()
                step = delta * step
                return val - delta / 2 - step, val + delta / 2 + step
            else:
                val, bottom, top = self.pg.value_at_cursor()
                delta = (top - bottom) * step
                return val - delta / 2, val + delta / 2

        # Setup
        if self.plot.lock_btn.isFlat():
            QTest.mouseClick(self.plot.lock_btn, Qt.MouseButton.LeftButton)

        y_step = 0.165
        x_step = 0.25

        self.assertIsNotNone(self.add_channels([35]))
        self.mouseClick_WidgetItem(self.channels[0])
        self.processEvents()

        self.pg.viewbox.menu.set_x_zoom_mode()
        self.pg.viewbox.menu.set_y_zoom_mode()

        # click con center
        QTest.mouseClick(
            self.plot.plot.viewport(),
            Qt.MouseButton.LeftButton,
            Qt.KeyboardModifier.NoModifier,
            QPoint(int(self.pg.width() / 2), int(self.pg.height() / 2)),
        )
        self.processEvents()

        # Events Without Pressed Shift
        expected_x_zoom_in_range = get_expected_result(-x_step, True)
        # Press "I"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["x_zoom_in"]))
        self.processEvents()
        x_zoom_in_range = self.pg.x_range

        expected_x_zoom_out_range = get_expected_result(x_step * 2, True)
        # Press "O"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["x_zoom_out"]))
        self.processEvents()
        x_zoom_out_range = self.pg.x_range

        # Events with pressed Shift
        expected_y_zoom_in_range = get_expected_result(1 / (1 + y_step), False)
        # Press "I"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["y_zoom_in"]))
        self.processEvents()
        y_zoom_in_range = self.pg.signals[0].y_range

        expected_y_zoom_out_range = get_expected_result(1 + y_step, False)
        # Press "O"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["y_zoom_out"]))
        self.processEvents()

        y_zoom_out_range = self.pg.signals[0].y_range
        # Evaluate
        delta = pow(10, -4)
        # Key Shift wasn't pressed
        self.assertAlmostEqual(x_zoom_in_range[0], expected_x_zoom_in_range[0], delta=delta)
        self.assertAlmostEqual(x_zoom_in_range[1], expected_x_zoom_in_range[1], delta=delta)
        self.assertAlmostEqual(x_zoom_out_range[0], expected_x_zoom_out_range[0], delta=delta)
        self.assertAlmostEqual(x_zoom_out_range[1], expected_x_zoom_out_range[1], delta=delta)
        # Key Shift was pressed
        self.assertAlmostEqual(y_zoom_in_range[0], expected_y_zoom_in_range[0], delta=delta)
        self.assertAlmostEqual(y_zoom_in_range[1], expected_y_zoom_in_range[1], delta=delta)
        self.assertAlmostEqual(y_zoom_out_range[0], expected_y_zoom_out_range[0], delta=delta)
        self.assertAlmostEqual(y_zoom_out_range[1], expected_y_zoom_out_range[1], delta=delta)

    def test_range_shortcut(self):
        """
        Test Scope:
            Check if Range Selection rectangle is painted over the plot.
        Events:
            - Press Key R for range selection
            - Move Cursors
            - Press Key R for range selection
        Evaluate:
            - Evaluate that two cursors are available
            - Evaluate that new rectangle with different color is present
            - Evaluate that sum of rectangle areas is same with rectangle of full black plot.
            - Evaluate that range selection disappears after pressing key R second time.
        """
        self.widget.set_cursor_options(False, False, 1, Pixmap.COLOR_CURSOR)
        # Save PixMap of clear plot
        clear_pixmap = self.pg.grab()
        self.assertTrue(Pixmap.is_black(clear_pixmap))

        # Get X position of Cursor
        cursors = Pixmap.cursors_x(clear_pixmap)
        # Evaluate that there is only one cursor
        self.assertEqual(1, len(cursors))

        # Press Key 'R' for range selection
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["range"]))
        self.processEvents(timeout=0.01)

        # Save PixMap of Range plot
        range_pixmap = self.pg.grab()
        self.assertFalse(Pixmap.is_black(range_pixmap))

        # Get X position of Cursors
        cursors = Pixmap.cursors_x(range_pixmap)
        # Evaluate that two cursors are available
        self.assertEqual(2, len(cursors))

        # Evaluate that new rectangle with different color is present
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=0,
                y=0,
                width=min(cursors) - 1,
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_RANGE,
                x=min(cursors) + 1,
                y=0,
                width=max(cursors),
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=max(cursors) + 1,
                y=0,
            )
        )

        # Move Cursors
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_right_1x"]))
        self.processEvents(timeout=0.01)
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_left_20x"]))
        self.processEvents(timeout=0.01)

        # Save PixMap of Range plot
        range_pixmap = self.pg.grab()
        self.assertFalse(Pixmap.is_black(range_pixmap))

        # Get X position of Cursors
        new_cursors = Pixmap.cursors_x(range_pixmap)
        # Evaluate that two cursors are available
        self.assertEqual(2, len(cursors))
        for c in cursors:
            self.assertNotIn(c, new_cursors, f"cursor {c} is the same")

        # Evaluate that new rectangle with different color is present
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=0,
                y=0,
                width=min(new_cursors) - 1,
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_RANGE,
                x=min(new_cursors) + 1,
                y=0,
                width=max(new_cursors),
            )
        )
        self.assertTrue(
            Pixmap.is_colored(
                pixmap=range_pixmap,
                color_name=Pixmap.COLOR_BACKGROUND,
                x=max(new_cursors) + 1,
                y=0,
            )
        )
        # Press Key 'R' for range selection
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["range"]))
        self.processEvents(timeout=0.01)

        # Save PixMap of clear plot
        clear_pixmap = self.pg.grab()
        self.assertTrue(Pixmap.is_black(clear_pixmap))

    def test_save_active_subplot_channels_shortcut(self):
        """
        Test Scope:
            Check if by pressing "Ctrl+S" is saved in the new measurement file only active channels.

        Events:
            - Open 'FileWidget' with valid measurement.
            - Select 3 signals and create a new plot
            - Deselect first channel from first plot
            - Press Key "Ctrl+S"
            - Open recently created measurement file in a new window.

        Evaluate:
            - Evaluate that object getSaveFileName() was called after pressing combination "Ctrl+S"
            - Evaluate that in measurement file is saved only active channels
        """
        # Setup
        file_path = os.path.join(self.test_workspace, "test_file.mf4")
        self.create_window(window_type="Plot", channels_indexes=(20, 21, 22))
        self.processEvents()
        second_plot_items = []
        iterator = QTreeWidgetItemIterator(self.widget.mdi_area.subWindowList()[1].widget().channel_selection)
        while item := iterator.value():
            second_plot_items.append(item.name)
            iterator += 1
        self.assertIsNotNone(self.add_channels([10, 11, 12, 13]))
        expected_channels = [channel.name for channel in self.channels]
        expected_channels.append("time")
        self.processEvents()

        # select all channels excluding the first one
        for _ in range(1, len(self.channels)):
            self.channels[_].setSelected(True)
        expected_channels.remove(self.channels[0].name)
        self.processEvents()

        with mock.patch("asammdf.gui.widgets.plot.QtWidgets.QFileDialog.getSaveFileName") as mo_getSaveFileName:
            mo_getSaveFileName.return_value = (file_path, "")
            # Press Ctrl+S
            QTest.keySequence(self.plot, QKeySequence(self.shortcuts["save_active_subplot_channels"]))
        # Evaluate
        mo_getSaveFileName.assert_called()

        # get waved file as MDF
        process_bus_logging = ("process_bus_logging", True)
        mdf_file = mdf.MDF(file_path, process_bus_logging=process_bus_logging)

        # Evaluate
        for name in expected_channels:
            self.assertIn(name, mdf_file.channels_db.keys())
        for name in second_plot_items:
            self.assertNotIn(name, mdf_file.channels_db.keys())
        mdf_file.close()

    def test_move_cursor_left__right_shortcuts(self):
        """
        Test Scope:
            Check that Arrow Keys: Left & Right ensure navigation on channels evolution.
            Ensure that navigation is working.
        Events:
            - Display one signal to plot
            - Click Right key specific number of times.
            - Click Left key specific number of times.
            - Click Ctrl+Right key specific number of times.
            - Click Ctrl+Left key specific number of times.
        Evaluate:
            - Evaluate values from selected channel value and cursor info, it must be equal to the expected values.
        """
        # Setup
        if self.plot.selected_channel_value_btn.isFlat():
            QTest.mouseClick(self.plot.selected_channel_value_btn, Qt.MouseButton.LeftButton)
        self.add_channels([37])
        ch = self.channels[0]
        # Number of times that specific key will be pressed
        right_clicks = 50
        ctrl_right_clicks = 20
        left_clicks = 30
        ctrl_left_clicks = 15
        ci = self.plot.cursor_info
        # Select channel
        self.mouseClick_WidgetItem(ch)
        self.pg.setFocus()
        self.processEvents(0.1)
        pos = 0
        cursor_prev = Pixmap.cursors_x(self.pg.grab())[0]
        c1_pos = self.pg.cursor1.getXPos()

        # Evaluate
        self.assertEqual(f"{ch.signal.samples[pos]} {ch.unit}", self.plot.selected_channel_value.text())
        self.assertEqual(f"{ci.name} = {round(ch.signal.timestamps[pos], ci.precision)}{ci.unit}", ci.text())
        self.assertEqual(ch.signal.timestamps[pos], self.pg.cursor1.getXPos())

        # Event
        for _ in range(right_clicks):
            QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_right_1x"]))
            self.processEvents(0.01)
        self.processEvents(0.5)
        pos += right_clicks
        cursor_now = Pixmap.cursors_x(self.pg.grab())[0]

        # Evaluate
        self.assertGreater(cursor_now, cursor_prev)
        self.assertEqual(f"{ch.signal.samples[pos]} {ch.unit}", self.plot.selected_channel_value.text())
        self.assertEqual(f"{ci.name} = {round(ch.signal.timestamps[pos], ci.precision)}{ci.unit}", ci.text())
        self.assertEqual(ch.signal.timestamps[pos], self.pg.cursor1.getXPos())

        # New setup
        pos -= left_clicks
        # Send Key strokes
        for _ in range(left_clicks):
            QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_left_1x"]))
            self.processEvents(0.01)
        self.processEvents(0.5)
        cursor_prev = cursor_now
        cursor_now = Pixmap.cursors_x(self.pg.grab())[0]

        # Evaluate
        self.assertLess(cursor_now, cursor_prev)
        self.assertEqual(f"{ch.signal.samples[pos]} {ch.unit}", self.plot.selected_channel_value.text())
        self.assertEqual(f"{ci.name} = {round(ch.signal.timestamps[pos], ci.precision)}{ci.unit}", ci.text())
        self.assertEqual(ch.signal.timestamps[pos], self.pg.cursor1.getXPos())

        # Send Key strokes
        for _ in range(ctrl_right_clicks):
            QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_right_20x"]))
            self.processEvents(0.01)
        self.processEvents(0.1)
        pos += 20 * ctrl_right_clicks
        cursor_prev = cursor_now
        cursor_now = Pixmap.cursors_x(self.pg.grab())[0]

        # Evaluate
        self.assertGreater(cursor_now, cursor_prev)
        self.assertEqual(f"{ch.signal.samples[pos]} {ch.unit}", self.plot.selected_channel_value.text())
        self.assertEqual(f"{ci.name} = {round(ch.signal.timestamps[pos], ci.precision)}{ci.unit}", ci.text())
        self.assertEqual(ch.signal.timestamps[pos], self.pg.cursor1.getXPos())

        # Send Key strokes
        for _ in range(ctrl_left_clicks):
            QTest.keySequence(self.pg, QKeySequence(self.shortcuts["move_cursor_left_20x"]))
            self.processEvents(0.01)
        self.processEvents(0.1)
        pos -= 20 * ctrl_left_clicks
        cursor_prev = cursor_now
        cursor_now = Pixmap.cursors_x(self.pg.grab())[0]

        # Evaluate
        self.assertLess(cursor_now, cursor_prev)
        self.assertEqual(f"{ch.signal.samples[pos]} {ch.unit}", self.plot.selected_channel_value.text())
        self.assertEqual(f"{ci.name} = {round(ch.signal.timestamps[pos], ci.precision)}{ci.unit}", ci.text())
        self.assertEqual(ch.signal.timestamps[pos], self.pg.cursor1.getXPos())

    @unittest.skipIf(sys.platform != "win32", "RuntimeError. C++ object <<ViewBoxWithCursor>> already deleted.")
    def test_shift_channels_shortcut(self):
        """
        Test Scope:
            Check that Shift + Arrow Keys ensure moving of selected channels.
        Events:
            - Create plot with 2 channels
            - Press key "S" to separate signals for better evaluation
            - Click on first channel
            - Press "Shift" key + arrow "Down" & "Left"
            - Click on second channel
            - Press "Shift" key + arrow "Up" & "Right"
        Evaluate:
            - Evaluate that first signal is shifted down & left after pressing combination "Shift+Down" & "Shift+Left"
            - Evaluate that second signal is shifted up & right after pressing combination "Shift+Up" & "Shift+Right"
        """
        if self.plot.lock_btn.isFlat():
            QTest.mouseClick(self.plot.lock_btn, Qt.MouseButton.LeftButton)
        self.add_channels([36, 37])
        channel_36 = self.channels[0]
        channel_37 = self.channels[1]

        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["stack_all"]))
        self.processEvents(0.01)

        # Zoom out
        x = round(self.plot.plot.width() / 2)
        y = round(self.plot.plot.height() / 2)
        QTest.mouseClick(self.pg.viewport(), Qt.MouseButton.LeftButton, Qt.KeyboardModifier.NoModifier, QPoint(x, y))
        self.wheel_action(self.pg.viewport(), x, y, -1)
        self.processEvents(0.1)

        # Find extremes of signals
        old_from_to_y_channel_36 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_36.color.name(), "y")
        old_from_to_x_channel_36 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_36.color.name(), "x")
        old_from_to_y_channel_37 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_37.color.name(), "y")
        old_from_to_x_channel_37 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_37.color.name(), "x")

        # Select first channel and move signal using commands Shift + PgDown/Down/Left
        self.mouseClick_WidgetItem(channel_36)
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_down_10x"]))
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_down_1x"]))
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_left"]))
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_left"]))
        self.avoid_blinking_issue(self.plot.channel_selection)
        self.processEvents()

        # Select second channel and move signal using commands Shift + PgUp/Up/Right
        self.mouseClick_WidgetItem(channel_37)
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_up_10x"]))
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_up_1x"]))
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_right"]))
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["shift_channels_right"]))
        self.avoid_blinking_issue(self.plot.channel_selection)
        self.processEvents()

        # Find new extremes
        new_from_to_y_channel_36 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_36.color.name(), "y")
        new_from_to_y_channel_37 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_37.color.name(), "y")
        new_from_to_x_channel_36 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_36.color.name(), "x")
        new_from_to_x_channel_37 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_37.color.name(), "x")

        # Evaluate
        self.assertLess(old_from_to_y_channel_36[0], new_from_to_y_channel_36[0])
        self.assertLess(old_from_to_y_channel_36[1], new_from_to_y_channel_36[1])
        self.assertGreater(old_from_to_x_channel_36[0], new_from_to_x_channel_36[0])
        self.assertGreater(old_from_to_x_channel_36[1], new_from_to_x_channel_36[1])

        self.assertGreater(old_from_to_y_channel_37[0], new_from_to_y_channel_37[0])
        self.assertGreater(old_from_to_y_channel_37[1], new_from_to_y_channel_37[1])
        self.assertLess(old_from_to_x_channel_37[0], new_from_to_x_channel_37[0])
        self.assertLess(old_from_to_x_channel_37[1], new_from_to_x_channel_37[1])

    def test_test_honeywell_shortcut(self):
        """
        Test Scope:
            Check if honeywell function is applied to signal after pressing key "H"

        Events:

            - Display 1 signal on plot
            - Select signal
            - Press "H"
            - Set window mode to full screen
            - Pres "H"

        Evaluate:
            - Evaluate the range of x-axis after pressing key "H", "honey range" must be respected
            - Evaluate the range of x-axis is same for maximized window
            - Evaluate the range of x-axis after pressing key "H" second time, "honey range" must be respected
        """

        def find_honey_range(plot):
            rect = plot.plotItem.vb.sceneBoundingRect()
            dpi = QGuiApplication.primaryScreen().physicalDotsPerInchX()
            dpc = dpi / 2.54  # from inch to cm
            physical_viewbox_width = (rect.width() - 5) / dpc  # cm
            return physical_viewbox_width * 0.1

        # Setup
        self.add_channels([35])
        expected_normal_screen_honey_range = find_honey_range(self.pg)
        # Press "H"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["honeywell"]))
        self.avoid_blinking_issue(self.plot.channel_selection)
        delta_normal_screen_x_range = self.pg.x_range[1] - self.pg.x_range[0]
        # Evaluate
        self.assertAlmostEqual(delta_normal_screen_x_range, expected_normal_screen_honey_range, delta=0.0001)

        # Minimize widget
        self.widget.setFixedSize(int(self.widget.width() * 0.9), int(self.widget.height() * 0.9))
        self.processEvents()
        # Evaluate
        self.assertAlmostEqual(self.pg.x_range[1] - self.pg.x_range[0], delta_normal_screen_x_range, delta=0.0001)

        expected_full_screen_honey_range = find_honey_range(self.pg)
        # Press "H"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["honeywell"]))
        self.avoid_blinking_issue(self.plot.channel_selection)
        delta_full_screen_x_range = self.pg.x_range[1] - self.pg.x_range[0]

        # Evaluate
        self.assertNotEqual(delta_full_screen_x_range, delta_normal_screen_x_range)
        self.assertAlmostEqual(delta_full_screen_x_range, expected_full_screen_honey_range, delta=0.0001)

    def test_home_shortcuts(self):
        """
        Check if the signal is fitted properly after pressing key "W".
        Events:
            - Create a plot window with 2 signals
            - Press key "I"
            - Press key "W"
        Evaluate:
            - Evaluate that there is at least one column with first signal color
            - Evaluate first and last columns where is first signal:
                > first column after pressing "I" is full black => signal colors are not there
                > signal is zoomed => is extended to left side => last column contain signal color
            - Evaluate that after pressing key "W", signal is displayed from first to last column
        """
        self.assertIsNotNone(self.add_channels([35]))
        channel_35 = self.channels[0]

        # check if the grid is available
        if not self.plot.hide_axes_btn.isFlat():
            QTest.mouseClick(self.plot.hide_axes_btn, Qt.MouseButton.LeftButton)

        # Press "W"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["home"]))
        self.processEvents(0.01)

        # search first and last column where is displayed first signal
        extremes_of_channel_35 = Pixmap.search_signal_extremes_by_ax(self.pg.grab(), channel_35.color.name(), ax="x")
        # Evaluate that there are extremes of first signal
        self.assertTrue(extremes_of_channel_35)
        # Press "I"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["x_zoom_in"]))
        self.processEvents(0.5)

        # save left and right pixel column
        x_left_column = self.pg.grab(QRect(extremes_of_channel_35[0], 0, 1, self.pg.height()))
        x_right_column = self.pg.grab(QRect(extremes_of_channel_35[1], 0, 1, self.pg.height()))
        self.assertTrue(Pixmap.is_black(x_left_column))
        self.assertTrue(Pixmap.has_color(x_right_column, channel_35.color.name()))

        # press "F"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["fit_all"]))
        # Press "W"
        QTest.keySequence(self.pg, QKeySequence(self.shortcuts["home"]))
        self.processEvents()
        # Select all columns from left to right
        for x in range(self.pg.height() - 1):
            column = self.pg.grab(QRect(x, 0, 1, self.pg.height()))
            if x < extremes_of_channel_35[0] - 1:
                self.assertTrue(Pixmap.is_black(column), f"column {x} is not black")
            elif extremes_of_channel_35[0] <= x <= extremes_of_channel_35[1]:
                self.assertTrue(
                    Pixmap.has_color(column, channel_35.color.name()),
                    f"column {x} doesn't have {channel_35.name} color",
                )
            else:
                self.assertTrue(Pixmap.is_black(column), f"column {x} is not black")

    def test_insert_computation_shortcut(self):
        """
        Test Scope:
            Check Insert key shortcut action

        Events:
            - Press Insert with preconditions:
                > There is no user defined functon
                > There is user defined functon, button Cancel or "X" was pressed
                > There is user defined function, button Apply was pressed

        Evaluate:
            - Evaluate that channel selection area is empty
        Evaluate (0):
            - Evaluate that warning message box was triggered after pressing key insert
            - Evaluate displayed warning message
            - Evaluate that channel selection area is empty
        Evaluate (1):
            - Evaluate that channel selection area is empty
            - Evaluate that DefineChannel object was called
        Evaluate (2):
            - Evaluate that there is one channel in channel selection area
            - Evaluate that the name of this channel is correct
            - Evaluate that DefineChannel object was called
        """
        # Evaluate precondition
        self.assertEqual(0, self.plot.channel_selection.topLevelItemCount())

        with self.subTest("_0_test_warning_no_user_function_defined"):
            warnings_msgs = [
                "Cannot add computed channel",
                "There is no user defined function. Create new function using the Functions Manger (F6)",
            ]
            # mock for warning message box
            with mock.patch("asammdf.gui.widgets.plot.MessageBox.warning") as mo_waring:
                # Press key Insert
                QTest.keySequence(self.plot.channel_selection, QKeySequence(self.shortcuts["insert_computation"]))
            # Evaluate
            self.assertEqual(0, self.plot.channel_selection.topLevelItemCount())
            mo_waring.assert_called()
            for w in warnings_msgs:
                self.assertIn(w, mo_waring.call_args.args)

        with self.subTest("_1_test_cancel_dlg_with_user_function_defined"):
            file_name = "test_insert_cfg.dspf"
            file_path = os.path.join(self.resource, file_name)
            self.load_display_file(file_path)
            self.plot = self.widget.mdi_area.subWindowList()[0].widget()
            with mock.patch("asammdf.gui.widgets.plot.DefineChannel") as mo_DefineChannel:
                # Press key Insert
                QTest.keySequence(self.plot.channel_selection, QKeySequence(self.shortcuts["insert_computation"]))

            # Evaluate
            self.assertEqual(1, self.plot.channel_selection.topLevelItemCount())
            mo_DefineChannel.assert_called()

        with self.subTest("_2_test_apply_dlg_with_user_function_defined"):
            file_name = "test_insert_cfg.dspf"
            file_path = os.path.join(self.resource, file_name)
            self.load_display_file(file_path)
            self.plot = self.widget.mdi_area.subWindowList()[0].widget()
            computed_channel = {
                "type": "channel",
                "common_axis": False,
                "individual_axis": False,
                "enabled": True,
                "mode": "phys",
                "fmt": "{:.3f}",
                "format": "phys",
                "precision": 3,
                "flags": 0,
                "ranges": [],
                "unit": "",
                "computed": True,
                "color": "#994380",
                "uuid": "525ad72a531a",
                "origin_uuid": "812d7b792168",
                "group_index": -1,
                "channel_index": -1,
                "name": self.id(),
                "computation": {
                    "args": {},
                    "type": "python_function",
                    "definition": "",
                    "channel_name": "Function_728d4a149b44",
                    "function": "Function1",
                    "channel_unit": "",
                    "channel_comment": "",
                    "triggering": "triggering_on_all",
                    "triggering_value": "all",
                    "computation_mode": "sample_by_sample",
                },
            }
            with mock.patch("asammdf.gui.widgets.plot.DefineChannel") as mo_DefineChannel:
                mo_DefineChannel.return_value.result = computed_channel
                # Press key Insert
                QTest.keySequence(self.plot.channel_selection, QKeySequence(self.shortcuts["insert_computation"]))

            # Evaluate
            self.assertEqual(2, self.plot.channel_selection.topLevelItemCount())
            self.assertEqual(self.plot.channel_selection.topLevelItem(1).name, self.id())
            mo_DefineChannel.assert_called()



#!/usr/bin/env python3
#
# This script generates constants for efficient computation of the gzip CRC-32.

import sys

# This is the generator polynomial G(x) of the gzip CRC-32, represented as an
# int using the natural mapping between bits and polynomial coefficients.
G = 0x104c11db7

# XOR (add) an iterable of polynomials.
def xor(iterable):
    res = 0
    for val in iterable:
        res ^= val
    return res

# Multiply two polynomials.
def clmul(a, b):
    return xor(a << i for i in range(b.bit_length()) if (b & (1 << i)) != 0)

# Polynomial division floor(a / b).
def div(a, b):
    q = 0
    while a.bit_length() >= b.bit_length():
        q ^= 1 << (a.bit_length() - b.bit_length())
        a ^= b << (a.bit_length() - b.bit_length())
    return q

# Reduce the polynomial 'a' modulo the polynomial 'b'.
def reduce(a, b):
    return a ^ clmul(div(a, b), b)

# Reverse the bits of a polynomial.
def bitreverse(poly, num_bits):
    return xor(1 << (num_bits - 1 - i) for i in range(num_bits)
               if (poly & (1 << i)) != 0)

# Compute x^d mod G.
def x_to_the_d(d):
    if d < G.bit_length() - 1:
        return 1 << d
    t = x_to_the_d(d//2)
    t = clmul(t, t)
    if d % 2 != 0:
        t <<= 1
    return reduce(t, G)

def gen_tables():
    print('/*')
    print(' * crc32_tables.h - data tables for CRC-32 computation')
    print(' *')
    print(' * THIS FILE WAS GENERATED BY gen-crc32-consts.py.  DO NOT EDIT.')
    print(' */')
    for n in [1, 8]:
        print('')
        print(f'static const u32 crc32_slice{n}_table[] MAYBE_UNUSED = {{')
        # The i'th table entry is the CRC-32 of the message consisting of byte
        # i % 256 followed by i // 256 zero bytes.
        polys = [bitreverse(i % 256, 8) << (32 + 8*(i//256)) for i in range(256 * n)]
        polys = [bitreverse(reduce(poly, G), 32) for poly in polys]
        for i in range(0, len(polys), 4):
            print(f'\t0x{polys[i+0]:08x}, 0x{polys[i+1]:08x}, 0x{polys[i+2]:08x}, 0x{polys[i+3]:08x},')
        print('};')

# Compute the constant multipliers needed for "folding" over various distances
# with the gzip CRC-32.  Each such multiplier is x^d mod G(x) for some distance
# d, in bits, over which the folding is occurring.
#
# Folding works as follows: let A(x) be a polynomial (possibly reduced partially
# or fully mod G(x)) for part of the message, and let B(x) be a polynomial
# (possibly reduced partially or fully mod G(x)) for a later part of the
# message.  The unreduced combined polynomial is A(x)*x^d + B(x), where d is the
# number of bits separating the two parts of the message plus len(B(x)).  Since
# mod G(x) can be applied at any point, x^d mod G(x) can be precomputed and used
# instead of x^d unreduced.  That allows the combined polynomial to be computed
# relatively easily in a partially-reduced form A(x)*(x^d mod G(x)) + B(x), with
# length max(len(A(x)) + 31, len(B(x))).  This does require doing a polynomial
# multiplication (carryless multiplication).
#
# "Folding" in this way can be used for the entire CRC computation except the
# final reduction to 32 bits; this works well when CPU support for carryless
# multiplication is available.  It can also be used to combine CRCs of different
# parts of the message that were computed using a different method.
#
# Note that the gzip CRC-32 uses bit-reversed polynomials.  I.e., the low order
# bits are really the high order polynomial coefficients.
def gen_multipliers():
    print('/*')
    print(' * crc32_multipliers.h - constants for CRC-32 folding')
    print(' *')
    print(' * THIS FILE WAS GENERATED BY gen-crc32-consts.py.  DO NOT EDIT.')
    print(' */')
    print('')

    # Compute the multipliers needed for CRC-32 folding with carryless
    # multiplication instructions that operate on the 64-bit halves of 128-bit
    # segments.  Using the terminology from earlier, for each 64-bit fold
    # len(A(x)) = 64, and len(B(x)) = 95 since a 64-bit polynomial multiplied by
    # a 32-bit one produces a 95-bit one.  When A(x) is the low order polynomial
    # half of a 128-bit segments (high order physical half), the separation
    # between the message parts is the total length of the 128-bit segments
    # separating the values.  When A(x) is the high order polynomial half, the
    # separation is 64 bits greater.
    for i in range(1, 33):
        sep_lo = 128 * (i - 1)
        sep_hi = sep_lo + 64
        len_B = 95
        for d in [sep_hi + len_B, # A(x) = high 64 polynomial bits (low 64 physical bits)
                  sep_lo + len_B # A(x) = low 64 polynomial bits (high 64 physical bits)
                  ]:
            poly = bitreverse(x_to_the_d(d), 32)
            print(f'#define CRC32_X{d}_MODG 0x{poly:08x} /* x^{d} mod G(x) */')
        print('')

    # Compute constants for the final 128 => 32 bit reduction.
    poly = bitreverse(div(1 << 95, G), 64)
    print(f'#define CRC32_BARRETT_CONSTANT_1 0x{poly:016x}ULL /* floor(x^95 / G(x)) */')
    poly = bitreverse(G, 33)
    print(f'#define CRC32_BARRETT_CONSTANT_2 0x{poly:016x}ULL /* G(x) */')

    # Compute multipliers for combining the CRCs of separate chunks.
    print('')
    num_chunks = 4
    table_len = 129
    min_chunk_len = 128
    print(f'#define CRC32_NUM_CHUNKS {num_chunks}')
    print(f'#define CRC32_MIN_VARIABLE_CHUNK_LEN {min_chunk_len}UL')
    print(f'#define CRC32_MAX_VARIABLE_CHUNK_LEN {(table_len-1) * min_chunk_len}UL')
    print('')
    print('/* Multipliers for implementations that use a variable chunk length */')
    print('static const u32 crc32_mults_for_chunklen[][CRC32_NUM_CHUNKS - 1] MAYBE_UNUSED = {')
    print('\t{ 0 /* unused row */ },')
    for i in range(1, table_len):
        chunk_len = i * min_chunk_len
        print(f'\t/* chunk_len={chunk_len} */')
        print('\t{ ', end='')
        for j in range(num_chunks - 1, 0, -1):
            d = (j * 8 * chunk_len) - 33
            poly = bitreverse(x_to_the_d(d), 32)
            print(f'0x{poly:08x} /* x^{d} mod G(x) */, ', end='')
        print('},')
    print('};')
    fixed_chunk_len = 32768
    print('')
    print('/* Multipliers for implementations that use a large fixed chunk length */')
    print(f'#define CRC32_FIXED_CHUNK_LEN {fixed_chunk_len}UL')
    for j in range(1, num_chunks):
        d = (j * 8 * fixed_chunk_len) - 33
        poly = bitreverse(x_to_the_d(d), 32)
        print(f'#define CRC32_FIXED_CHUNK_MULT_{j} 0x{poly:08x} /* x^{d} mod G(x) */')

with open('lib/crc32_tables.h', 'w') as f:
    sys.stdout = f
    gen_tables()
with open('lib/crc32_multipliers.h', 'w') as f:
    sys.stdout = f
    gen_multipliers()



#!/usr/bin/env python
from pathlib import Path
from unittest import mock

import numpy as np

from test.asammdf.gui.test_base import OpenMDF
from test.asammdf.gui.widgets.test_BaseBatchWidget import TestBatchWidget

# Note: If it's possible and make sense, use self.subTests
# to avoid initializing widgets multiple times and consume time.


class TestPushButtons(TestBatchWidget):
    default_test_file = "ASAP2_Demo_V171.mf4"
    class_test_file = "test_batch.mf4"

    def setUp(self):
        """
        Events
            - Open 'BatchWidget' with 2 valid measurement.
            - Go to Tab: "Stack"
        """
        super().setUp()
        self.copy_mdf_files_to_workspace()

        # Setup
        self.measurement_file_1 = Path(self.test_workspace, self.default_test_file)
        self.measurement_file_2 = Path(self.test_workspace, self.class_test_file)
        self.saved_file = Path(self.test_workspace, self.id().split(".")[-1] + ".mf4")

        self.setUpBatchWidget(measurement_files=[str(self.measurement_file_1), str(self.measurement_file_2)])
        self.widget.aspects.setCurrentIndex(self.stack_aspect)

    def test_PushButton_Stack(self):
        """
        Events:
            - Press PushButton "Stack"
            - Save measurement file

        Evaluate:
            - New file is created
            - All channels samples and timestamps from both input files can be found in selected output file
        """
        # Event
        with mock.patch("asammdf.gui.widgets.batch.QtWidgets.QFileDialog.getSaveFileName") as mo_getSaveFileName:
            mo_getSaveFileName.return_value = str(self.saved_file), ""
            self.mouse_click_on_btn_with_progress(self.widget.stack_btn)

        # Evaluate
        self.assertTrue(self.saved_file.exists())

        # Evaluate saved file
        with (
            OpenMDF(self.saved_file) as new_mdf_file,
            OpenMDF(self.measurement_file_1) as mdf_test_file_0,
            OpenMDF(self.measurement_file_2) as mdf_test_file_1,
        ):
            for channel in mdf_test_file_1.iter_channels():
                new_file_channel = new_mdf_file.get(channel.name)
                self.assertTrue(np.array_equal(channel.samples, new_file_channel.samples))

            for channel in mdf_test_file_0.iter_channels():
                new_file_channel = new_mdf_file.get(channel.name)
                self.assertTrue(np.array_equal(channel.samples, new_file_channel.samples))



from PySide6.QtCore import QMetaObject
from PySide6.QtUiTools import QUiLoader


class UiLoader(QUiLoader):
    def __init__(self, base_instance):
        QUiLoader.__init__(self, base_instance)
        self.base_instance = base_instance

    def createWidget(self, class_name, parent=None, name=""):
        if parent is None and self.base_instance:
            return self.base_instance
        else:
            # create a new widget for child widgets
            widget = QUiLoader.createWidget(self, class_name, parent, name)
            if self.base_instance:
                setattr(self.base_instance, name, widget)
            return widget


def load_ui(ui_file, base_instance=None):
    loader = UiLoader(base_instance)
    widget = loader.load(ui_file)
    QMetaObject.connectSlotsByName(widget)
    return widget
